"""Compact and full diffractive optical models."""
import copy
import numpy as np
import multiprocessing
import matplotlib.pyplot as plt
from . import jacobians
import falco
from falco import check
import proper
import logging
log = logging.getLogger(__name__)


def full(mp, modvar, isNorm=True):
    """
    Truth model used to generate images in simulation.

    Truth model used to generate images in simulation. Can include
    aberrations/errors that are unknown to the estimator and controller. This
    function is the wrapper for full models of any coronagraph type.

    Parameters
    ----------
    mp : ModelParameters
        Structure containing optical model parameters
    modvar : ModelVariables
        Structure containing temporary optical model variables
    isNorm : bool
        If False, return an unnormalized image. If True, return a
        normalized image with the currently stored norm value.

    Returns
    -------
    Eout : numpy ndarray
        2-D electric field in final focal plane
    """
    if type(mp) is not falco.config.ModelParameters:
        raise TypeError('Input "mp" must be of type ModelParameters')

    if hasattr(modvar, 'sbpIndex'):
        normFac = mp.Fend.full.I00[modvar.sbpIndex, modvar.wpsbpIndex]
        # Value to normalize the PSF. Set to 0 when finding the norm factor

    # Optional Keyword arguments
    if not isNorm:
        normFac = 0

    # Set the wavelength
    if(hasattr(modvar, 'wvl')):  # For FALCO or standalone use of full model
        wvl = modvar.wvl
    elif(hasattr(modvar, 'sbpIndex')):  # For use in FALCO
        wvl = mp.full.lambdasMat[modvar.sbpIndex,modvar.wpsbpIndex]
    else:
        log.warning('full: Need to specify a value or indices for a wavelength.')  

    """ Input E-fields """
    # Set the point source as the exoplanet or the star
    if modvar.whichSource.lower() == 'exoplanet': # Don't include tip/tilt jitter for planet wavefront since the effect is minor
        # The planet does not move in sky angle, so the actual tip/tilt angle needs to scale inversely with wavelength.
        #planetAmp = np.sqrt(mp.c_planet);  # Scale the E field to the correct contrast
        #planetPhase = (-1)*(2*np.pi*(mp.x_planet*mp.P2.full.XsDL + mp.y_planet*mp.P2.full.YsDL));
        #Ein = planetAmp*exp(1j*planetPhase*mp.lambda0/wvl)*mp.P1.full.E[:,:,modvar.wpsbpIndex,modvar.sbpIndex];
        pass
    elif modvar.whichSource.lower() == 'offaxis': # Use for throughput calculations 
        TTphase = (-1.)*(2*np.pi*(modvar.x_offset*mp.P2.full.XsDL + modvar.y_offset*mp.P2.full.YsDL));
        Ett = np.exp(1j*TTphase*mp.lambda0/wvl)
        Ein = Ett*np.squeeze(mp.P1.full.E[:,:,modvar.wpsbpIndex,modvar.sbpIndex]) 

    else:  # Default to using the starlight
        Ein = np.squeeze(mp.P1.full.E[:,:,modvar.wpsbpIndex,modvar.sbpIndex])  

    # Shift the source off-axis to compute the intensity normalization value.
    #  This replaces the previous way of taking the FPM out in the optical model.
    if normFac == 0:
        source_x_offset = mp.source_x_offset_norm # source offset in lambda0/D for normalization
        source_y_offset = mp.source_y_offset_norm # source offset in lambda0/D for normalization
        TTphase = (-1)*(2*np.pi*(source_x_offset*mp.P2.full.XsDL + source_y_offset*mp.P2.full.YsDL));
        Ett = np.exp(1j*TTphase*mp.lambda0/wvl)
        Ein = Ett*np.squeeze(mp.P1.full.E[:, :, modvar.wpsbpIndex,
                                          modvar.sbpIndex])

    # Apply a Zernike (in amplitude) at input pupil if specified
    if not (hasattr(modvar, 'zernIndex')):
        modvar.zernIndex = 1
    if not modvar.zernIndex == 1:
        indsZnoll = modvar.zernIndex  # Just send in 1 Zernike mode
        zernMat = np.squeeze(falco.zern.gen_norm_zern_maps(mp.P1.full.Nbeam,
                                                           mp.centering,
                                                           indsZnoll))
        zernMat = falco.util.pad_crop(zernMat, mp.P1.full.Narr)
        Ein = Ein*zernMat*(2*np.pi/wvl)*mp.jac.Zcoef[mp.jac.zerns ==
                                                     modvar.zernIndex]

    # Pre-compute the FPM first for HLC as mp.FPM.mask
    mp.FPM = falco.config.Object()
    if mp.layout.lower() == 'fourier' or mp.layout.lower() == 'proper':
        pass
    elif mp.layout.lower() == 'fpm_scale':
        pass

    # Select which optical layout's full model to use.
    if mp.layout.lower() == 'fourier':
        Eout = full_Fourier(mp, wvl, Ein, normFac)

    elif mp.layout.lower() == 'fpm_scale':  # FPM scales with wavelength
        Eout = full_Fourier(mp, wvl, Ein, normFac, flagScaleFPM=True)

    elif mp.layout.lower() == 'proper':

        optval = copy.copy(vars(mp.full))

        if any(mp.dm_ind == 1):
            optval['use_dm1'] = True
            optval['dm1'] = mp.dm1.V*mp.dm1.VtoH + mp.full.dm1FlatMap

        if any(mp.dm_ind == 2):
            optval['use_dm2'] = True
            optval['dm2'] = mp.dm2.V*mp.dm2.VtoH + mp.full.dm2FlatMap

        if normFac == 0:
            optval['xoffset'] = -mp.source_x_offset_norm
            optval['yoffset'] = -mp.source_y_offset_norm
            if mp.coro.upper() in ('VORTEX', 'VC', 'AVC'):
                optval['use_fpm'] = False
                optval['xoffset'] = 0
                optval['yoffset'] = 0
                pass

        # wavelength needs to be in microns instead of meters for PROPER
        [Eout, sampling_m] = proper.prop_run(mp.full.prescription, wvl*1e6,
                                             mp.P1.full.Narr, QUIET=True,
                                             PASSVALUE=optval)
        if not normFac == 0:
            Eout = Eout/np.sqrt(normFac)

        del optval

    elif mp.layout.lower() == 'wfirst_phaseb_proper':
        optval = copy.copy(vars(mp.full))
        optval['use_dm1'] = True
        optval['use_dm2'] = True
        optval['dm1_m'] = mp.dm1.V*mp.dm1.VtoH + mp.full.dm1.flatmap
        optval['dm2_m'] = mp.dm2.V*mp.dm2.VtoH + mp.full.dm2.flatmap
        if normFac == 0:
            optval['source_x_offset'] = -mp.source_x_offset_norm
            optval['source_y_offset'] = -mp.source_y_offset_norm

        [Eout, sampling] = \
            proper.prop_run('wfirst_phaseb', wvl*1e6,
                            int(2**falco.util.nextpow2(mp.Fend.Nxi)),
                            QUIET=True, PASSVALUE=optval)
        Eout = falco.util.pad_crop(Eout, (mp.Fend.Nxi, mp.Fend.Nxi))

        if not normFac == 0:
            Eout = Eout/np.sqrt(normFac)

    return Eout


def full_Fourier(mp, wvl, Ein, normFac, flagScaleFPM=False):
    """
    Truth model with a simple layout used to generate images in simulation.

    Truth model used to generate images in simulation. Can include
    aberrations/errors that are unknown to the estimator and controller.
    This function uses the simplest model (FTs except for angular spectrum
    between DMs) for several coronagraph types.

    Parameters
    ----------
    mp : ModelParameters
        Structure containing optical model parameters
    wvl : Wavelength
        Scalar value for the wavelength of the light in meters
    Ein : Electric field input
        2-D electric field in the input pupil
    normFac : Normalization factor
        Scalar value of the PSF peak normalization factor to apply to the whole
        image.
    flagScaleFPM : bool, optional
        Whether to scale the diameter of the FPM inversely with wavelength.

    Returns
    -------
    Eout : numpy ndarray
        2-D electric field in final focal plane

    """
    check.is_bool(flagScaleFPM, 'flagScaleFPM')

    mirrorFac = 2  # Phase change is twice the DM surface height in reflection
    NdmPad = int(mp.full.NdmPad)
    if flagScaleFPM:
        fpmScaleFac = wvl/mp.lambda0
    else:
        fpmScaleFac = 1.0

    """ Masks and DM surfaces """
    if any(mp.dm_ind == 1):
        DM1surf = falco.dm.gen_surf_from_act(mp.dm1, mp.dm1.dx, NdmPad)
#        try:
#            DM1surf = falco.util.pad_crop(mp.dm1.surfM, NdmPad)
#        except AttributeError:  # No surfM parameter exists, create DM surface
#            DM1surf = falco.dm.gen_surf_from_act(mp.dm1, mp.dm1.dx, NdmPad)
    else:
        DM1surf = np.zeros((NdmPad, NdmPad))

    if any(mp.dm_ind == 2):
        DM2surf = falco.dm.gen_surf_from_act(mp.dm2, mp.dm2.dx, NdmPad)
#        try:
#            DM2surf = falco.util.pad_crop(mp.dm2.surfM, NdmPad)
#        except AttributeError:  # No surfM parameter exists, create DM surface
#            DM2surf = falco.dm.gen_surf_from_act(mp.dm2, mp.dm2.dx, NdmPad)
    else:
        DM2surf = np.zeros((NdmPad, NdmPad))

    pupil = falco.util.pad_crop(mp.P1.full.mask, NdmPad)
    Ein = falco.util.pad_crop(Ein, NdmPad)

    if mp.flagDM1stop:
        DM1stop = falco.util.pad_crop(mp.dm1.full.mask, NdmPad)
    else:
        DM1stop = np.ones((NdmPad, NdmPad))

    if mp.flagDM2stop:
        DM2stop = falco.util.pad_crop(mp.dm2.full.mask, NdmPad)
    else:
        DM2stop = np.ones((NdmPad, NdmPad))

    if(mp.flagDMwfe):
        pass
        #if(any(mp.dm_ind==1));  Edm1WFE = exp(2*np.pi*1j/wvl*pad_crop(mp.dm1.wfe,NdmPad,'extrapval',0)); else; Edm1WFE = ones(NdmPad); end
        #if(any(mp.dm_ind==2));  Edm2WFE = exp(2*np.pi*1j/wvl*pad_crop(mp.dm2.wfe,NdmPad,'extrapval',0)); else; Edm2WFE = ones(NdmPad); end
    else:
        Edm1WFE = np.ones((NdmPad, NdmPad))
        Edm2WFE = np.ones((NdmPad, NdmPad))


    """ Propagation: entrance pupil, 2 DMs, (optional) apodizer, FPM, LS, and final focal plane """

    # Define pupil P1 and Propagate to pupil P2
    EP1 = pupil*Ein # E-field at pupil plane P1
    EP2 = falco.prop.relay(EP1, mp.Nrelay1to2, mp.centering) # Forward propagate to the next pupil plane (P2) by rotating 180 degrees mp.Nrelay1to2 times.

    # Propagate from P2 to DM1, and apply DM1 surface and aperture stop
    if not ( abs(mp.d_P2_dm1)==0 ): 
        Edm1 = falco.prop.ptp(EP2, mp.P2.full.dx*NdmPad,wvl, mp.d_P2_dm1) 
    else: 
        Edm1 = EP2   # E-field arriving at DM1
    Edm1b = Edm1*Edm1WFE*DM1stop*np.exp(mirrorFac*2*np.pi*1j*DM1surf/wvl) # E-field leaving DM1

    # Propagate from DM1 to DM2, and apply DM2 surface and aperture stop
    Edm2 = falco.prop.ptp(Edm1b, mp.P2.full.dx*NdmPad,wvl, mp.d_dm1_dm2)
    Edm2 *= Edm2WFE*DM2stop*np.exp(mirrorFac*2*np.pi*1j*DM2surf/wvl)

    # Back-propagate to pupil P2
    if(mp.d_P2_dm1 + mp.d_dm1_dm2 == 0):
        EP2eff = Edm2 # Do nothing if zero distance
    else:
        EP2eff = falco.prop.ptp(Edm2, mp.P2.full.dx*NdmPad,wvl,-1*(mp.d_dm1_dm2 + mp.d_P2_dm1)) # Back propagate to pupil P2

    # Re-image to pupil P3
    EP3 = falco.prop.relay(EP2eff, mp.Nrelay2to3, mp.centering)

    # Apply the apodizer mask (if there is one)
    if(mp.flagApod):
        EP3 = mp.P3.full.mask*falco.util.pad_crop(EP3, mp.P3.full.Narr)

    # Propagations Specific to the Coronagraph Type
    if mp.coro.upper() in ('LC', 'APLC', 'RODDIER'):
        # MFT from apodizer plane to FPM (i.e., P3 to F3)
        EF3inc = falco.prop.mft_p2f(EP3, mp.fl, wvl, mp.P2.full.dx,
                                    fpmScaleFac*mp.F3.full.dxi, mp.F3.full.Nxi,
                                    fpmScaleFac*mp.F3.full.deta,
                                    mp.F3.full.Neta, mp.centering)
        # Apply (1-FPM) for Babinet's principle later
        if mp.coro.upper() == 'RODDIER':
            FPM = mp.F3.full.ampMask*np.exp(1j*2*np.pi/wvl*(mp.F3.n(wvl)-1) *
                                            mp.F3.t*mp.F3.full.mask.phzSupport)
            EF3 = (1.-FPM)*EF3inc  # Apply (1-FPM) for Babinet's princ. later
        else:
            EF3 = (1.-mp.F3.full.ampMask)*EF3inc
        #  Use Babinet's principle at the Lyot plane.
        EP4noFPM = falco.prop.relay(EP3, mp.Nrelay3to4, mp.centering)
        #  MFT from FPM to Lyot Plane (i.e., F3 to P4)
        EP4subtrahend = falco.prop.mft_f2p(EF3, mp.fl, wvl,
                                           fpmScaleFac*mp.F3.full.dxi,
                                           mp.F3.full.deta,
                                           fpmScaleFac*mp.P4.full.dx,
                                           mp.P4.full.Narr, mp.centering)
        # Babinet's principle at P4
        EP4 = falco.util.pad_crop(EP4noFPM, mp.P4.full.Narr) - EP4subtrahend

    elif(mp.coro.upper() == 'SPLC' or mp.coro.upper() == 'FLC'):
        # MFT from apodizer plane to FPM (i.e., P3 to F3)
        EF3inc = falco.prop.mft_p2f(EP3, mp.fl, wvl, mp.P2.full.dx,
                                    fpmScaleFac*mp.F3.full.dxi, mp.F3.full.Nxi,
                                    fpmScaleFac*mp.F3.full.deta,
                                    mp.F3.full.Neta, mp.centering)
        EF3 = mp.F3.full.ampMask * EF3inc

        # MFT from FPM to Lyot Plane (i.e., F3 to P4)
        EP4 = falco.prop.mft_f2p(EF3, mp.fl, wvl, fpmScaleFac*mp.F3.full.dxi,
                                 fpmScaleFac*mp.F3.full.deta, mp.P4.full.dx,
                                 mp.P4.full.Narr, mp.centering)
        EP4 = falco.prop.relay(EP4, mp.Nrelay3to4-1, mp.centering)

    elif mp.coro.upper() in ('VORTEX', 'VC', 'AVC'):
        # Get FPM charge
        if isinstance(mp.F3.VortexCharge, np.ndarray):
            # Passing an array for mp.F3.VortexCharge with
            # corresponding wavelengths mp.F3.VortexCharge_lambdas
            # represents a chromatic vortex FPM
            if mp.F3.VortexCharge.size == 1:
                charge = mp.F3.VortexCharge
            else:
                np.interp(wvl, mp.F3.VortexCharge_lambdas, mp.F3.VortexCharge,
                          'linear', 'extrap')

        elif isinstance(mp.F3.VortexCharge, (int, float)):
            # single value indicates fully achromatic mask
            charge = mp.F3.VortexCharge
        else:
            raise TypeError("mp.F3.VortexCharge must be an int, float, or\
                            numpy ndarray.")
            pass
        EP4 = falco.prop.mft_p2v2p(EP3, charge, mp.P1.full.Nbeam/2., 0.3, 5)
        EP4 = falco.util.pad_crop(EP4, mp.P4.full.Narr)

    else:
        log.warning('The chosen coronagraph type is not included yet.')
        raise ValueError("Value of mp.coro not recognized.")
        pass

    # Remove FPM completely if normalization value being found for vortex
    if normFac == 0:
        if mp.coro.upper() in ('VORTEX', 'VC', 'AVC'):
            EP4 = falco.prop.relay(EP3, mp.Nrelay3to4, mp.centering)
            EP4 = falco.util.pad_crop(EP4, mp.P4.full.Narr)
            pass
        pass

    """ Back to common propagation any coronagraph type """
    # Apply the (cropped-down) Lyot stop
    EP4 *= mp.P4.full.croppedMask

    # MFT from Lyot Stop to final focal plane (i.e., P4 to Fend)
    EP4 = falco.prop.relay(EP4, mp.NrelayFend, mp.centering)
    EFend = falco.prop.mft_p2f(EP4, mp.fl, wvl, mp.P4.full.dx, mp.Fend.dxi,
                               mp.Fend.Nxi, mp.Fend.deta, mp.Fend.Neta,
                               mp.centering)

    # Don't apply FPM if normalization value is being found
    if normFac == 0:
        Eout = EFend  # Don't normalize if normalization value is being found
    else:
        Eout = EFend/np.sqrt(normFac)  # Apply normalization

    return Eout


def compact(mp, modvar, isNorm=True, isEvalMode=False):
    """
    Simplified (aka compact) model used by estimator and controller.
    
    Simplified (aka compact) model used by estimator and controller. Does not include 
    unknown aberrations of the full, "truth" model. This function is the wrapper for 
    compact models of any coronagraph type.

    Parameters
    ----------
    mp : ModelParameters
        Structure containing optical model parameters
    modvar : ModelVariables
        Structure containing temporary optical model variables
    isNorm : bool
        If False, return an unnormalized image. If True, return a 
        normalized image with the currently stored norm value.
    isEvalMode : bool
        If set, uses a higher resolution in the focal plane for 
        measuring performance metrics such as throughput.

    Returns
    -------
    Eout : array_like
        2-D electric field in final focal plane

    """
    if type(mp) is not falco.config.ModelParameters:
        raise TypeError('Input "mp" must be of type ModelParameters')

    # Set default values of input parameters
    normFac = mp.Fend.compact.I00[modvar.sbpIndex]  # Value to normalize PSF.
    flagEval = False  # use a different res at final focal plane for eval

    # Optional Keyword arguments
    if not isNorm:
        normFac = 0.
    if isEvalMode:
        flagEval = True

    # Normalization factor for compact evaluation model
    if(isNorm and isEvalMode):
        normFac = mp.Fend.eval.I00[modvar.sbpIndex]
        # Value to normalize the PSF. Set to 0 when finding the norm factor

    # Set the wavelength
    if hasattr(modvar, 'wvl'):
        wvl = modvar.wvl
    else:
        wvl = mp.sbp_centers[modvar.sbpIndex]

    """ Input E-fields """

    # Include the tip/tilt in the input wavefront
#    if(hasattr(mp,'ttx')):
#         %--Scale by wvl/lambda0 because ttx and tty are in lambda0/D
#         x_offset = mp.ttx(modvar.ttIndex)*(mp.lambda0/wvl);
#         y_offset = mp.tty(modvar.ttIndex)*(mp.lambda0/wvl);
#
#         TTphase = (-1)*(2*np.pi*(x_offset*mp.P2.compact.XsDL +
#                    y_offset*mp.P2.compact.YsDL));
#         Ett = exp(1j*TTphase*mp.lambda0/wvl);
#         Ein = Ett.*mp.P1.compact.E(:,:,modvar.sbpIndex)
    if modvar.whichSource.lower() == 'offaxis':  # Use for throughput calc
        TTphase = (-1)*(2*np.pi*(modvar.x_offset*mp.P2.compact.XsDL +
                                 modvar.y_offset*mp.P2.compact.YsDL))
        Ett = np.exp(1j*TTphase*mp.lambda0/wvl)
        Ein = Ett*mp.P1.compact.E[:, :, modvar.sbpIndex]
    else:  # Backward compatible with code without tip/tilt offsets in Jacobian
        Ein = mp.P1.compact.E[:, :, modvar.sbpIndex]

    # Shift the source off-axis to compute the intensity normalization value.
    # This replaces the previous way of taking the FPM out in optical model.
    if normFac == 0:
        # source offset in lambda0/D for normalization
        source_x_offset = mp.source_x_offset_norm
        source_y_offset = mp.source_y_offset_norm
        TTphase = (-1.)*(2*np.pi*(source_x_offset*mp.P2.compact.XsDL +
                                  source_y_offset*mp.P2.compact.YsDL))
        Ett = np.exp(1j*TTphase*mp.lambda0/wvl)
        Ein = Ett*mp.P1.compact.E[:, :, modvar.sbpIndex]

    # Apply a Zernike (in amplitude) at input pupil if specified
    if not hasattr(modvar, 'zernIndex'):
        modvar.zernIndex = 1

    # Only used for Zernike sensitivity control, which requires the perfect
    # E-field of the differential Zernike term.
    if not (modvar.zernIndex == 1):
        indsZnoll = modvar.zernIndex  # Just send in 1 Zernike mode
        zernMat = np.squeeze(falco.zern.gen_norm_zern_maps(mp.P1.compact.Nbeam,
                                                           mp.centering,
                                                           indsZnoll))
        zernMat = falco.util.pad_crop(zernMat, mp.P1.compact.Narr)
        Ein = Ein*zernMat*(2*np.pi*1j/wvl)*mp.jac.Zcoef[mp.jac.zerns ==
                                                        modvar.zernIndex]

    # Define what the complex-valued FPM is if the coro is some type of HLC.
    mp.FPM = falco.config.Object()
    if mp.layout.lower() == 'fourier':
        if mp.coro.upper() in ('HLC',):
            mp.FPM.mask = falco.thinfilm.gen_hlc_fpm_complex_trans_mat(mp, modvar.sbpIndex, modvar.wpsbpIndex, 'compact')
    elif mp.layout.lower() == 'wfirst_phaseb_proper':
        if mp.coro.upper() in ('HLC',):
            mp.FPM.mask = mp.compact.FPMcube[:, :, modvar.sbpIndex]

    # Select which optical layout's compact model to use and get E-field
    if mp.layout.lower() == 'fourier' or mp.layout.lower() == 'proper':
        Eout = compact_general(mp, wvl, Ein, normFac, flagEval)

    elif mp.layout.lower() == 'fpm_scale':
        if mp.coro.upper() in ('HLC', 'SPLC'):
            Eout = compact_general(mp, wvl, Ein, normFac, flagEval,
                                   flagScaleFPM=True)

    elif mp.layout.lower() == 'wfirst_phaseb_proper':
        if mp.coro.upper() == 'HLC':
            Eout = compact_general(mp, wvl, Ein, normFac, flagEval,
                                   flagScaleFPM=True)
        elif 'SP' in mp.coro.upper():
            Eout = compact_general(mp, wvl, Ein, normFac, flagEval)
    return Eout


def compact_general(mp, wvl, Ein, normFac, flagEval, flagScaleFPM=False):
    """
    Compact model with a general-purpose optical layout.

    Used by estimator and controller.

    Simplified (aka compact) model used by estimator and controller. Does not
    include unknown aberrations of the full, "truth" model. This has a general
    optical layout that should work for most applications.

    Parameters
    ----------
    mp : ModelParameters
        Structure containing optical model parameters
    wvl : float
        Wavelength of light [meters]
    Ein : numpy ndarray
        2-D input electric field
    normFac : float
        Intensity normalization factor
    flagEval : bool
        Whether to use a higher resolution in final image plane for evaluation.
    flagScaleFPM : bool, optional
        Whether to scale the diameter of the FPM inversely with wavelength.

    Returns
    -------
    Eout : numpy ndarray
        2-D electric field in final focal plane

    """
    if type(mp) is not falco.config.ModelParameters:
        raise TypeError('Input "mp" must be of type ModelParameters')
    check.is_bool(flagEval, 'flagEval')
    check.is_bool(flagScaleFPM, 'flagScaleFPM')

    mirrorFac = 2.  # Phase change is twice the DM surface height.
    NdmPad = int(mp.compact.NdmPad)
    if flagScaleFPM:
        fpmScaleFac = wvl/mp.lambda0
    else:
        fpmScaleFac = 1.0

    if(flagEval):  # Higher resolution at final focal plane for eval
        dxi = mp.Fend.eval.dxi
        Nxi = mp.Fend.eval.Nxi
        deta = mp.Fend.eval.deta
        Neta = mp.Fend.eval.Neta
    else:  # Otherwise use the detector resolution
        dxi = mp.Fend.dxi
        Nxi = mp.Fend.Nxi
        deta = mp.Fend.deta
        Neta = mp.Fend.Neta

    """ Masks and DM surfaces """
    # ompute the DM surfaces for the current DM commands
    if any(mp.dm_ind == 1):
        DM1surf = falco.dm.gen_surf_from_act(mp.dm1, mp.dm1.compact.dx,
                                              NdmPad)
    else:
        DM1surf = np.zeros((NdmPad, NdmPad))
    if any(mp.dm_ind == 2):
        DM2surf = falco.dm.gen_surf_from_act(mp.dm2, mp.dm2.compact.dx,
                                              NdmPad)
    else:
        DM2surf = np.zeros((NdmPad, NdmPad))

    pupil = falco.util.pad_crop(mp.P1.compact.mask, NdmPad)
    Ein = falco.util.pad_crop(Ein, NdmPad)

    if(mp.flagDM1stop):
        DM1stop = falco.util.pad_crop(mp.dm1.compact.mask, NdmPad)
    else:
        DM1stop = np.ones((NdmPad, NdmPad))
    if(mp.flagDM2stop):
        DM2stop = falco.util.pad_crop(mp.dm2.compact.mask, NdmPad)
    else:
        DM2stop = np.ones((NdmPad, NdmPad))

    if(mp.useGPU):
        log.warning('GPU support not yet implemented. Proceeding without GPU.')

    # This block is for BMC surface error testing
    if mp.flagDMwfe:
        if any(mp.dm_ind == 1):
            Edm1WFE = np.exp(2*np.pi*1j/wvl *
                             falco.util.pad_crop(mp.dm1.compact.wfe,
                                                  NdmPad, 'extrapval', 0))
        else:
            Edm1WFE = np.ones((NdmPad, NdmPad))
        if any(mp.dm_ind == 2):
            Edm2WFE = np.exp(2*np.pi*1j/wvl *
                             falco.util.pad_crop(mp.dm2.compact.wfe,
                                                  NdmPad, 'extrapval', 0))
        else:
            Edm2WFE = np.ones((NdmPad, NdmPad))
    else:
        Edm1WFE = np.ones((NdmPad, NdmPad))
        Edm2WFE = np.ones((NdmPad, NdmPad))

    """Propagation"""

    # Define pupil P1 and Propagate to pupil P2
    EP1 = pupil*Ein  # E-field at pupil plane P1
    EP2 = falco.prop.relay(EP1, mp.Nrelay1to2, mp.centering)

    # Propagate from P2 to DM1, and apply DM1 surface and aperture stop
    if not (abs(mp.d_P2_dm1) == 0):  # E-field arriving at DM1
        Edm1 = falco.prop.ptp(EP2, mp.P2.compact.dx*NdmPad, wvl, mp.d_P2_dm1)
    else:
        Edm1 = EP2
    # E-field leaving DM1
    Edm1b = Edm1*Edm1WFE*DM1stop*np.exp(mirrorFac*2*np.pi*1j*DM1surf/wvl)

    # Propagate from DM1 to DM2, and apply DM2 surface and aperture stop
    Edm2 = falco.prop.ptp(Edm1b, mp.P2.compact.dx*NdmPad, wvl, mp.d_dm1_dm2)
    Edm2 *= Edm2WFE*DM2stop*np.exp(mirrorFac*2*np.pi*1j*DM2surf/wvl)

    # Back-propagate to pupil P2
    if(mp.d_P2_dm1 + mp.d_dm1_dm2 == 0):
        EP2eff = Edm2
    else:
        EP2eff = falco.prop.ptp(Edm2, mp.P2.compact.dx*NdmPad, wvl, -1 *
                                (mp.d_dm1_dm2 + mp.d_P2_dm1))

    # Re-image to pupil P3
    EP3 = falco.prop.relay(EP2eff, mp.Nrelay2to3, mp.centering)

    # Apply apodizer mask.
    if(mp.flagApod):
        EP3 = mp.P3.compact.mask*falco.util.pad_crop(EP3, mp.P3.compact.Narr)


# Remove FPM completely if normalization value is being found for vortex
    if normFac == 0:
        if mp.coro.upper() in ('VORTEX', 'VC', 'AVC', 'LC', 'APLC'):
            EP4 = falco.prop.relay(EP3, mp.Nrelay3to4, mp.centering)
            EP4 = falco.util.pad_crop(EP4, mp.P4.compact.Narr)
            
        elif mp.coro.upper() in ('HLC'):
            EP4noFPM = falco.prop.relay(EP3, mp.Nrelay3to4, mp.centering)
            EP4 = falco.util.pad_crop(EP4noFPM, mp.P4.compact.Narr)
            FPM = mp.FPM.mask  # Complex transmission of the FPM
            transOuterFPM = FPM[0, 0]  # Complex trans of points outside FPM
            EP4 *= transOuterFPM
    
    else:

        """  Select propagation based on coronagraph type   """
        if mp.coro.upper() in ('LC', 'APLC', 'HLC'):
            # MFT from SP to FPM (i.e., P3 to F3)
            # E-field incident upon the FPM
            EF3inc = falco.prop.mft_p2f(EP3, mp.fl, wvl, mp.P2.compact.dx,
                                        fpmScaleFac*mp.F3.compact.dxi,
                                        mp.F3.compact.Nxi,
                                        fpmScaleFac*mp.F3.compact.deta,
                                        mp.F3.compact.Neta, mp.centering)
            # Apply (1-FPM) for Babinet's principle later
            if mp.coro.upper() == 'HLC':
                FPM = mp.FPM.mask  # Complex transmission of the FPM
                transOuterFPM = FPM[0, 0]  # Complex trans of points outside FPM
                EF3 = (transOuterFPM - FPM)*EF3inc
                # transOuterFPM instead of 1 because of the complex transmission of
                # the glass as well as the arbitrary phase shift.
            else:
                EF3 = (1. - mp.F3.compact.ampMask)*EF3inc
            # Use Babinet's principle at the Lyot plane.
            EP4noFPM = falco.prop.relay(EP3, mp.Nrelay3to4, mp.centering)
            EP4noFPM = falco.util.pad_crop(EP4noFPM, mp.P4.compact.Narr)
            if mp.coro.upper() == 'HLC':
                EP4noFPM = transOuterFPM*EP4noFPM
            # MFT from FPM to Lyot Plane (i.e., F3 to P4)
            # Subtrahend term for Babinet's principle
            EP4sub = falco.prop.mft_f2p(EF3, mp.fl, wvl,
                                        fpmScaleFac*mp.F3.compact.dxi,
                                        fpmScaleFac*mp.F3.compact.deta,
                                        mp.P4.compact.dx, mp.P4.compact.Narr,
                                        mp.centering)
            EP4subRelay = falco.prop.relay(EP4sub, mp.Nrelay3to4-1, mp.centering)
            # Babinet's principle at P4
            EP4 = (EP4noFPM-EP4subRelay)
    
        elif mp.coro.upper() == 'FLC' or mp.coro.upper() == 'SPLC':
            # MFT from SP to FPM (i.e., P3 to F3)
            # E-field incident upon the FPM
            EF3inc = falco.prop.mft_p2f(EP3, mp.fl, wvl, mp.P2.compact.dx,
                                        mp.F3.compact.dxi, mp.F3.compact.Nxi,
                                        mp.F3.compact.deta, mp.F3.compact.Neta,
                                        mp.centering)
    
            # Apply FPM
            EF3 = mp.F3.compact.ampMask * EF3inc
    
            # MFT from FPM to Lyot Plane (i.e., F3 to P4)
            EP4 = falco.prop.mft_f2p(EF3, mp.fl, wvl, mp.F3.compact.dxi,
                                     mp.F3.compact.deta, mp.P4.compact.dx,
                                     mp.P4.compact.Narr, mp.centering)
            EP4 = falco.prop.relay(EP4, mp.Nrelay3to4-1, mp.centering)
    
        elif mp.coro.upper() in ('VORTEX', 'VC', 'AVC'):
    
            # Get FPM charge
            if isinstance(mp.F3.VortexCharge, np.ndarray):
                # Passing an array for mp.F3.VortexCharge with
                # corresponding wavelengths mp.F3.VortexCharge_lambdas
                # represents a chromatic vortex FPM
                if mp.F3.VortexCharge.size == 1:
                    charge = mp.F3.VortexCharge
                else:
                    np.interp(wvl, mp.F3.VortexCharge_lambdas, mp.F3.VortexCharge,
                              'linear', 'extrap')
    
            elif isinstance(mp.F3.VortexCharge, (int, float)):
                # single value indicates fully achromatic mask
                charge = mp.F3.VortexCharge
            else:
                raise TypeError("mp.F3.VortexCharge must be int, float or numpy\
                                ndarray.")
                pass
            EP4 = falco.prop.mft_p2v2p(EP3, charge, mp.P1.compact.Nbeam/2., 0.3, 5)
            EP4 = falco.util.pad_crop(EP4, mp.P4.compact.Narr)
    
        else:
            raise ValueError("Value of mp.coro not recognized.")
            pass

    """  Back to common propagation any coronagraph type   """
    # Apply the Lyot stop
    EP4 = mp.P4.compact.croppedMask*EP4

    # MFT to camera
    EP4 = falco.prop.relay(EP4, mp.NrelayFend, mp.centering)
    EFend = falco.prop.mft_p2f(EP4, mp.fl, wvl, mp.P4.compact.dx, dxi, Nxi,
                               deta, Neta, mp.centering)

    # Don't apply FPM if normalization value is being found
    if normFac == 0:
        Eout = EFend  # Don't normalize if normalization value is being found
    else:
        Eout = EFend/np.sqrt(normFac)  # Apply normalization

    return Eout


def jacobian(mp):
    """
    Compute the control Jacobian used for EFC.

    Wrapper function for the function model_Jacobian_middle_layer, which gets
    the DM response matrix, aka the control Jacobian for each specified DM.

    Parameters
    ----------
    mp : ModelParameters
        Structure containing optical model parameters

    Returns
    -------
    jacStruct : ModelParameters
        Structure containing the Jacobians for each specified DM.

    """
    jacStruct = falco.config.Object()  # Initialize the new structure

    # Pre-compute the DM surfaces to save time
    NdmPad = int(mp.compact.NdmPad)
    if any(mp.dm_ind == 1):
        mp.dm1.compact.surfM = falco.dm.gen_surf_from_act(mp.dm1,
                                                           mp.dm1.compact.dx,
                                                           NdmPad)
    else:
        mp.dm1.compact.surfM = np.zeros((NdmPad, NdmPad))
    if any(mp.dm_ind == 2):
        mp.dm2.compact.surfM = falco.dm.gen_surf_from_act(mp.dm2,
                                                           mp.dm2.compact.dx,
                                                           NdmPad)
    else:
        mp.dm2.compact.surfM = np.zeros((NdmPad, NdmPad))

    # Initialize the Jacobians for each DM
    if any(mp.dm_ind == 1):
        jacStruct.G1 = np.zeros((mp.Fend.corr.Npix, mp.dm1.Nele, mp.jac.Nmode),
                                dtype=np.complex)
    if any(mp.dm_ind == 2):
        jacStruct.G2 = np.zeros((mp.Fend.corr.Npix, mp.dm2.Nele, mp.jac.Nmode),
                                dtype=np.complex)
    if any(mp.dm_ind == 8):
        jacStruct.G8 = np.zeros((mp.Fend.corr.Npix, mp.dm8.Nele, mp.jac.Nmode),
                                dtype=np.complex)
    if any(mp.dm_ind == 9):
        jacStruct.G9 = np.zeros((mp.Fend.corr.Npix, mp.dm9.Nele, mp.jac.Nmode),
                                dtype=np.complex)

    # Calculate the Jacobian in parallel or serial
    if(mp.flagMultiproc):
        print('Computing control Jacobian matrices in parallel...', end='')
        pool = multiprocessing.Pool(processes=mp.Nthreads)

        with falco.util.TicToc():
#            results_order = [pool.apply(_func_Jac_ordering, args=(im,idm)) for im,idm in zip(*map(np.ravel, np.meshgrid(np.arange(mp.jac.Nmode,dtype=int),mp.dm_ind))) ]        
            results_order = [(im,idm) for idm in mp.dm_ind for im in np.arange(mp.jac.Nmode,dtype=int)] # Use for assigning parts of the Jacobian list to the correct DM and mode
            results = [pool.apply_async(_jac_middle_layer, args=(mp,im,idm)) for im,idm in zip(*map(np.ravel, np.meshgrid(np.arange(mp.jac.Nmode,dtype=int),mp.dm_ind))) ]
            results_Jac = [p.get() for p in results] # All the Jacobians in a list
            pool.close()
            pool.join()

            # Reorder Jacobian by mode and DM from the list
            for ii in range(mp.jac.Nmode*mp.dm_ind.size):
                im = results_order[ii][0]
                idm = results_order[ii][1]
                if idm == 1:
                    jacStruct.G1[:, :, im] = results_Jac[ii]
                if idm == 2:
                    jacStruct.G2[:, :, im] = results_Jac[ii]

            print('done.')

    else:
        print('Computing control Jacobian matrices in serial:\n  ', end='')
        with falco.util.TicToc():
            for im in range(mp.jac.Nmode):
                if(any(mp.dm_ind==1)):
                    print('mode%ddm%d...' % (im,1),end='')
                    jacStruct.G1[:,:,im] = _jac_middle_layer(mp, im, 1)
                if(any(mp.dm_ind==2)):
                    print('mode%ddm%d...' % (im,2),end='')
                    jacStruct.G2[:,:,im] = _jac_middle_layer(mp, im, 2)
            print('done.')

    ## TIED ACTUATORS

    # Handle tied actuators by adding the 2nd actuator's Jacobian column to
    #the first actuator's column, and then zeroing out the 2nd actuator's column.
    if any(mp.dm_ind == 1):
        # Update the sets of tied actuators
        mp.dm1 = falco.dm.enforce_constraints(mp.dm1)
        for ti in range(mp.dm1.tied.shape[0]):
            Index1all = mp.dm1.tied[ti, 0]  # Index of first tied actuator in whole actuator set. 
            Index2all = mp.dm1.tied[ti, 1]  # Index of second tied actuator in whole actuator set. 
            Index1subset = np.nonzero(mp.dm1.act_ele == Index1all)[0] # Index of first tied actuator in subset of used actuators. 
            Index2subset = np.nonzero(mp.dm1.act_ele == Index2all)[0] # Index of second tied actuator in subset of used actuators. 
            jacStruct.G1[:, Index1subset,:] += jacStruct.G1[:, Index2subset, :] # adding the 2nd actuators Jacobian column to the first actuator's column
            jacStruct.G1[:, Index2subset,:] = 0*jacStruct.G1[:, Index2subset, :] # zero out the 2nd actuator's column.

    if any(mp.dm_ind == 2):
        mp.dm2 = falco.dm.enforce_constraints(mp.dm2) # Update the sets of tied actuators
        for ti in range(mp.dm2.tied.shape[0]):
            Index1all = mp.dm2.tied[ti,0] # Index of first tied actuator in whole actuator set. 
            Index2all = mp.dm2.tied[ti,1] # Index of second tied actuator in whole actuator set. 
            Index1subset = np.nonzero(mp.dm2.act_ele==Index1all)[0] # Index of first tied actuator in subset of used actuators. 
            Index2subset = np.nonzero(mp.dm2.act_ele==Index2all)[0] # Index of second tied actuator in subset of used actuators. 
            jacStruct.G2[:, Index1subset, :] += jacStruct.G2[:, Index2subset, :] # adding the 2nd actuators Jacobian column to the first actuator's column
            jacStruct.G2[:, Index2subset, :] = 0*jacStruct.G2[:, Index2subset, :] # zero out the 2nd actuator's column.

    return jacStruct


def _func_Jac_ordering(im, idm):
    """Order modes for parallelized Jacobian calculation."""
    return (im, idm)


def _jac_middle_layer(mp, im, idm):
    """
    Select which optical layout's Jacobian model to use and get E-field.

    Parameters
    ----------
    mp : ModelParameters
        Structure containing optical model parameters

    Returns
    -------
    jacMode : numpy ndarray
        Complex-valued, 2-D array containing the Jacobian for the specified DM.

    """
    if mp.layout.lower() in ('fourier', 'proper'):
        if mp.coro.upper() in ('LC', 'APLC', 'FLC', 'SPLC'):
            jacMode = jacobians.lyot(mp, im, idm)
        elif mp.coro.upper() in ('VC', 'AVC', 'VORTEX'):
            jacMode = jacobians.vortex(mp, im, idm)
    elif mp.layout.lower() == 'wfirst_phaseb_proper':
        if mp.coro.upper() in ('HLC', 'SPC', 'SPLC'):
            jacMode = jacobians.lyot(mp, im, idm)
        else:
            raise ValueError('%s not recognized as value for mp.coro' %
                             mp.coro)
    else:
        raise ValueError('mp.layout.lower not recognized')

    return jacMode



"""FALCO utilities."""
import time
import numpy as np
import itertools
import math
from . import check


class TicToc(object):
    """Class for timing."""

    def __init__(self, name=None):
        """Initialize."""
        self.name = name

    def __enter__(self):
        """Start timer."""
        self.tstart = time.time()

    def __exit__(self, type, value, traceback):
        """End timer."""
        if self.name:
            print('[%s]\t' % self.name, end='')
        print('Elapsed: %s' % (time.time() - self.tstart))


def cart2pol(x, y):
    """
    Convert Cartesian coordinate(s) into polar coordinate(s).

    Parameters
    ----------
    x : float or numpy.ndarray
        x-axis coordinate(s)
    y : float or numpy.ndarray
        y-axis coordinate(s)

    Returns
    -------
    rho : float or numpy.ndarray
        radial coordinate(s)
    theta : float or numpy.ndarray
        azimuthal coordinate(s)
    """
    if(type(x) == np.ndarray and type(y) == np.ndarray):
        if not x.shape == y.shape:
            raise ValueError('The two inputs must have the same shape.')

    rho = np.sqrt(x**2 + y**2)
    theta = np.arctan2(y, x)

    return(rho, theta)


def sind(thetaDeg):
    """
    Compute the sine of the input given in degrees.

    Parameters
    ----------
    thetaDeg : float or numpy.ndarray
        Angle in degrees

    Returns
    -------
    float or numpy.ndarray
        sine of the input value
    """
    return math.sin(math.radians(thetaDeg))


def cosd(thetaDeg):
    """
    Compute the cosine of the input given in degrees.

    Parameters
    ----------
    thetaDeg : float or numpy.ndarray
        Angle in degrees

    Returns
    -------
    float or numpy.ndarray
        cosine of the input value
    """
    return math.cos(math.radians(thetaDeg))


def nextpow2(N):
    """Return exponent for the smallest power of 2 that satisfies 2^p ≥ |N|."""
    check.real_scalar(N, 'N', TypeError)
    if N == 0:
        p = int(0)
    else:
        p = int(np.ceil(np.log2(np.abs(N))))
    
    return p


def ceil_even(x_in):
    """
    Compute the next highest even integer above the input.

    Parameters
    ----------
    x_in : float
        Scalar value

    Returns
    -------
    x_out : integer
        Even-valued integer
    """
    check.real_scalar(x_in, 'x_in', TypeError)
    return int(2 * np.ceil(0.5 * x_in))


def ceil_odd(x_in):
    """
    Compute the next highest odd integer above the input.

    Parameters
    ----------
    x_in : float
        Scalar value

    Returns
    -------
    x_out : integer
        Odd-valued integer
    """
    check.real_scalar(x_in, 'x_in', TypeError)
    x_out = int(np.ceil(x_in))
    if x_out % 2 == 0:
        x_out += 1

    return x_out


def pad_crop(arrayIn, outsize, extrapval=0):
    """
    Insert a 2D array into another array, centered and zero-padded.

    Given an array and a tuple/list denoting the dimensions of a second array,
    this places the smaller array in the center of the larger array and
    returns that larger array.  The smaller array will be zero-padded to the
    size of the larger.  If ``arrayIn.shape`` is larger than ``outsize`` in
    either dimension, this dimension will be truncated.

    If both sizes of a dimension are even or odd, the array will be centered
    in that dimension.  If the input size is even and the output size is odd,
    the smaller array will be shifted one element toward the start of the
    array.  If the input is odd and the output is even, the smaller array will
    be shifted one element toward the end of the array.  This sequence is
    intended to ensure transitivity, so several ``pad_crop()`` calls can be
    chained in no particular order without changing the final result.

    The output array will be of the same type as the input array. It will be
    a copy even if the arrays are the same size.

    Parameters
    ----------
     arrayIn: numpy ndarray
         input array to be padded or cropped
     outsize: array_like, int
         A positive integer or 2-element tuple/list/ndarray of positive
         integers giving dimensions of output array. If outsize is an int, the
         output has square dimensions of (outsize, outsize)

    Returns
    -------
     arrayOut: numpy ndarray
        an ndarray of the same size as ``outsize`` and type as ``arrayIn``

    """
    check.twoD_array(arrayIn, 'arrayIn', TypeError)
    check.real_scalar(extrapval, 'extrapval', TypeError)
    sh0 = arrayIn.shape    # np.shape(arrayIn) #
    int_types = (int, np.integer)  # Int check support
    if isinstance(outsize, int_types):
        sh1 = (outsize, outsize)
    else:
        sh1 = outsize

    try:
        if len(sh1) != 2:
            raise TypeError('Output dimensions must have 2 elements')
        if (not isinstance(sh1[0], int)) or (not isinstance(sh1[1], int)):
            raise TypeError('Output dimensions must be integers')
        if (sh1[0] <= 0) or (sh1[1] <= 0):
            raise TypeError('Output dimensions must be positive ' + 'integers')
    except TypeError:
        raise TypeError('outsize must be an iterable')

    arrayOut = extrapval * np.ones(sh1, dtype=arrayIn.dtype)

    xneg = min(sh0[1]//2, sh1[1]//2)
    xpos = min(sh0[1] - sh0[1]//2, sh1[1] - sh1[1]//2)
    yneg = min(sh0[0]//2, sh1[0]//2)
    ypos = min(sh0[0] - sh0[0]//2, sh1[0] - sh1[0]//2)

    slice0 = (slice(sh0[0]//2-yneg, sh0[0]//2+ypos),
              slice(sh0[1]//2-xneg, sh0[1]//2+xpos))
    slice1 = (slice(sh1[0]//2-yneg, sh1[0]//2+ypos),
              slice(sh1[1]//2-xneg, sh1[1]//2+xpos))

    arrayOut[slice1] = arrayIn[slice0]

    return arrayOut


def allcomb(*args, **kwargs):
    """
    Compute the Cartesian product of a series of iterables.

    Compute the Cartesian product of a series of iterables, i.e. the list
    consisting of all n-tuples formed by choosing one element from each of the
    n inputs. The output list will have length (P1 x P2 x ... x PN), where
    P1, P2, ..., PN are the lengths of the N input lists.

    Examples
    --------
        allcomb([1, 3, 5], [-3, 8], [0, 1]) % numerical input:
            [(1, -3, 0), (1, -3, 1), (1, 8, 0), (1, 8, 1), (3, -3, 0),
             (3, -3, 1), (3, 8, 0),(3, 8, 1), (5, -3, 0), (5, -3, 1),
             (5, 8, 0), (5, 8, 1)]

        allcomb('abc','XY') % character arrays
            [('a', 'X'), ('a', 'Y'), ('b', 'X'), ('b', 'Y'), ('c', 'X'),
             ('c', 'Y')]

        allcomb('xy', [65, 66]) % a combination
            [('x', 65), ('x', 66), ('y', 65), ('y', 66)]
            # a 4-by-2 character array

    Parameters
    ----------
        An arbitrary long series of iterables.  May be of different lengths and
        types.

    Returns
    -------
    list of tuple
        Cartesian product of input lists (explained above)
    """
    return list(itertools.product(*args))


def _spec_arg(k, kwargs, v):
    """
    Specify default argument for class constructors created from .mat files.

    Used in autogenerated classes like ModelParameters.

    Parameters
    ----------
    k : string
        Name of variable whose value will be assigned

    kwargs : dict
        Dictionary of keyword arguments, which may or may not specify a value
        for k.  If the "mat_struct" field exists, then the value of k will be
        obtained by accessing the corresponding value inside the given MATLAB
        struct via kwargs["mat_struct"].

    v : any
        Default value for the variable k.

    Returns
    -------
    The value to initialze the class with.
    """
    if k in kwargs:
        return kwargs[k]

    elif "mat_struct" in kwargs:
        return eval("kwargs[\"mat_struct\"]." + k)

    else:
        return v


def broadcast(axis):
    """
    Use numpy array broadcasting in two dimensions.

    Use numpy array broadcasting to return two views of the input axis that
    behave like a row vector (x) and a column vector (y), and which can be used
    to build memory-efficient coordinate grids without using meshgrid.

    Given an axis with length N, the naive approach using meshgrid requires
    building two NxN arrays, and combining them (e.g. to obtain a radial
    coordinate grid) produces a third NxN array. Using array broadcasting, one
    only needs to create two separate views into the original Nx1 vector to
    create the final NxN array.

    Parameters
    ----------
    axis : array_like
        1D coordinate axis

    Returns
    -------
    array_like, array_like
        Two views into axis that behave like a row and column vector,
        respectively.

    """
    x = axis[None, :]
    y = axis[:, None]
    return x, y


def azimuthal_grid(axis, xStretch=1., yStretch=1.):
    """
    Compute a memory-efficient radial grid using array broadcasting.

    Parameters
    ----------
    axis : array_like
        1D coordinate axis

    Returns
    -------
    array_like
        2D grid with azimuthal coordinates generated by axis
    """
    check.oneD_array(axis, 'axis', TypeError)
    check.real_scalar(xStretch, 'xStretch', TypeError)
    check.real_scalar(yStretch, 'yStretch', TypeError)
    x, y = broadcast(axis)
    return np.arctan2(y/yStretch, x/xStretch)


def radial_grid(axis, xStretch=1., yStretch=1.):
    """
    Compute a memory-efficient radial grid using array broadcasting.

    Parameters
    ----------
    axis : array_like
        1D coordinate axis

    Returns
    -------
    array_like
        2D grid with radial coordinates generated by axis
    """
    check.oneD_array(axis, 'axis', TypeError)
    check.real_scalar(xStretch, 'xStretch', TypeError)
    check.real_scalar(yStretch, 'yStretch', TypeError)

    x, y = broadcast(axis)
    return np.sqrt((x/xStretch)**2 + (y/yStretch)**2)


def radial_grid_squared(axis, xStretch=1., yStretch=1.):
    """
    Compute a memory-efficient squared radial grid using array broadcasting.

    Parameters
    ----------
    axis : array_like
        1D coordinate axis

    Returns
    -------
    array_like
        2D grid with squared radial coordinates generated by axis
    """
    check.oneD_array(axis, 'axis', TypeError)
    check.real_scalar(xStretch, 'xStretch', TypeError)
    check.real_scalar(yStretch, 'yStretch', TypeError)

    x, y = broadcast(axis)
    return (x/xStretch)**2 + (y/yStretch)**2


def create_axis(N, step, centering='pixel'):
    """
    Create a one-dimensional coordinate axis with a given size and step size.

    Can be constructed to follow either the FFT (pixel-centered) interpixel-
    centered convention, which differ by half a pixel for even-sized arrays.
    For odd-sized arrays, both values of centering put the center on the center
    pixel.

    Parameters
    ----------
    N : int
        Number of pixels in output axis
    step : float
        Physical step size between axis elements
    centering : 'pixel' or 'interpixel'
        Centering of the coordinates in the array.  Note that if N is odd, the
        result will be pixel-centered regardless of the value of this keyword.

    Returns
    -------
    array_like
        The output coordinate axis
    """
    check.positive_scalar_integer(N, 'N', TypeError)
    check.real_positive_scalar(step, 'step', TypeError)
    check.centering(centering)

    axis = np.arange(-N // 2, N // 2, dtype=np.float64) * step
    even = not N % 2  # Even number of samples?

    if even and (centering == 'interpixel'):
        # Inter-pixel-centering onlyif the number of samples is even
        axis += 0.5 * step

    return axis


def offcenter_crop(arrayIn, centerRow, centerCol, nRowOut, nColOut):
    """
    Crop a 2-D array to be centered at the specified pixel.

    This function crops a 2-D array about the center pixel specified by
    centerRow and centerCol. The input array can be
    rectangular with even or odd side lengths. The output will is rectangular
    with dimensions nRowOut, nColOut. If the output array includes
    regions outside the input array, those pixels are included
    and set to zero. If the specified cropping region is fully outside the
    input array, then the output is all zeros.

    The center pixel of an odd-sized array is the array
    center, and the center pixel of an even-sized array follows the FFT
    center pixel convention.

    Parameters
    ----------
    arrayIn : array_like
        2-D input array
    centerRow, centerCol : float or int
        Indices of the pixel to be used as the output array's center.
        Floating point values are rounded to the nearest integer.
        Convention in this function is that y is the first axis.
        Values can be negative and/or lie outside the input array.
    nRowOut, nColOut : int
        Height and width of the 2-D output array in pixels.

    Returns
    -------
    recentered_image : numpy ndarray
        2-D square array

    Notes
    -----
    All alignment units are in detector pixels.
    """
    check.twoD_array(arrayIn, 'arrayIn', TypeError)
    check.real_scalar(centerRow, 'centerRow', TypeError)
    check.real_scalar(centerCol, 'centerCol', TypeError)
    check.positive_scalar_integer(nRowOut, 'nRowOut', TypeError)
    check.positive_scalar_integer(nColOut, 'nColOut', TypeError)

    [nRowIn, nColIn] = arrayIn.shape
    centerRow = int(np.round(centerRow))
    centerCol = int(np.round(centerCol))

    # Compute how much to pad the array in y (if any)
    rowPadPre = 0
    rowPadPost = 0
    if np.ceil(-nRowOut/2.) + centerRow < 0:
        rowPadPre = np.abs(np.ceil(-nRowOut/2.) + centerRow)
    if np.ceil(nRowOut/2.) + centerRow > (nRowIn - 1):
        rowPadPost = np.ceil(nRowOut/2.) + centerRow - \
          (nRowIn)
    y_pad = int(np.max((rowPadPre, rowPadPost)))

    # Compute how much to pad the array in x (if any)
    colPadPre = 0
    colPadPost = 0
    if np.ceil(-nColOut/2.) + centerCol < 0:
        colPadPre = np.abs(np.ceil(-nColOut/2.) + centerCol)
    if np.ceil(nColOut/2.) + centerCol > (nColIn - 1):
        colPadPost = np.ceil(nColOut/2.) + centerCol - \
                     (nColIn)
    x_pad = int(np.max((colPadPre, colPadPost)))

    arrayPadded = pad_crop(arrayIn, (nRowIn+2*y_pad,
                                      nColIn+2*x_pad))

    centerCol += x_pad
    centerRow += y_pad

    # Buffer needed to keep output array correct size
    if nRowOut % 2 == 1:
        rowBuffer = 1
    else:
        rowBuffer = 0
    if nColOut % 2 == 1:
        colBuffer = 1
    else:
        colBuffer = 0

    arrayOut = arrayPadded[
       centerRow-nRowOut//2:centerRow+nRowOut//2 + rowBuffer,
       centerCol-nColOut//2:centerCol+nColOut//2 + colBuffer]

    return arrayOut
    

def bin_downsample(Ain, dsfac):
    """
    Downsample an array by binning.
    
    Parameters
    ----------
    Ain : 2-D array
        The matrix to be downsampled
    dsfac : int
        Downsampling factor for the matrix
    
    Returns
    -------
    Aout : 2-D array
        Downsampled array
    """
    # Error checks on inputs
    check.twoD_array(Ain, 'Ain', ValueError)
    check.positive_scalar_integer(dsfac, 'dsfac', ValueError)
    
    # Array Sizes
    ny0, nx0 = Ain.shape
    if (nx0 % dsfac != 0) or (ny0 % dsfac != 0):
        ValueError('The size of Ain must be divisible by dsfac.')

    nx1 = int(nx0/dsfac)
    ny1 = int(ny0/dsfac)
    
    # Bin and average values from the high-res array into the low-res array
    Aout = np.zeros((ny1, nx1))
    for ix in range(nx1):
        for iy in range(ny1):
            Aout[iy, ix] = np.sum(Ain[dsfac*iy:dsfac*(iy+1),
                                      dsfac*ix:dsfac*(ix+1)])/dsfac/dsfac
            
    return Aout



# import copy
from numpy import inf
# import numpy as np
# import falco.mask
# import falco.util
from falco.util import _spec_arg
# from falco.config import DeformableMirrorParameters
# import collections
# from falco import model

class Object(object):
    pass

class ModelParameters:
#     class _base_compact:
#         def __init__(self, **kwargs):
#             self.dummy = _spec_arg("dummy", kwargs, 1)
# 
#     class _base_full:
#         def __init__(self, **kwargs):
#             self.dummy = _spec_arg("dummy", kwargs, 1)

    class _base_dm1:
        def __init__(self, **kwargs):
            self.dummy = _spec_arg("dummy", kwargs, 1)
            self.Dstop = _spec_arg("Dstop", kwargs, 0.048)
            self.dx= _spec_arg("dx", kwargs, 0)
            self.NdmPad= _spec_arg("NdmPad", kwargs, 0)

    class _base_dm2:
        def __init__(self, **kwargs):
            self.dummy = _spec_arg("dummy", kwargs, 1)
            self.Dstop = _spec_arg("Dstop", kwargs, 0)
            self.dx= _spec_arg("dx", kwargs, 0.048)
            self.NdmPad= _spec_arg("NdmPad", kwargs, 0)
    
    class _base_dm8:
        def __init__(self, **kwargs):
            self.dummy = _spec_arg("dummy", kwargs, 1)
            
    class _base_dm9:
        def __init__(self, **kwargs):
            self.dummy = _spec_arg("dummy", kwargs, 1)

    class _base_P2:
        def __init__(self, **kwargs):
            self.dummy = _spec_arg("dummy", kwargs, 1)
            self.D = _spec_arg("D", kwargs, 0.0463)

#     class _base_P3:
#         def __init__(self, **kwargs):
#             self.dummy = _spec_arg("dummy", kwargs, 1)
#             self.D = _spec_arg("D", kwargs, 0.0463)
                
    class _base_P1:
        class _base_compact:
            def __init__(self, **kwargs):
                self.dummy = _spec_arg("dummy", kwargs, 1)
                self.Nbeam = _spec_arg("Nbeam", kwargs, 324)

        class _base_full:
            def __init__(self, **kwargs):
                self.dummy = _spec_arg("dummy", kwargs, 1)
                self.Nbeam = _spec_arg("Nbeam", kwargs, 324)

        def __init__(self, **kwargs):
            self.compact = _spec_arg("compact", kwargs, self._base_compact())
            self.full = _spec_arg("full", kwargs, self._base_full())
#             self.D = _spec_arg("D", kwargs, 2.3631)

    class _base_P3:
        class _base_compact:
            def __init__(self, **kwargs):
                self.dummy = _spec_arg("dummy", kwargs, 1)
#                 self.Nbeam = _spec_arg("Nbeam", kwargs, 324)

        class _base_full:
            def __init__(self, **kwargs):
                self.dummy = _spec_arg("dummy", kwargs, 1)
        
        def __init__(self, **kwargs):
            self.compact = _spec_arg("compact", kwargs, self._base_compact())
            self.full = _spec_arg("full", kwargs, self._base_full())

    class _base_P4:
        class _base_compact:
            def __init__(self, **kwargs):
                self.dummy = _spec_arg("dummy", kwargs, 1)
#                 self.Nbeam = _spec_arg("Nbeam", kwargs, 324)

        class _base_full:
            def __init__(self, **kwargs):
                self.dummy = _spec_arg("dummy", kwargs, 1)
#                 self.Nbeam = _spec_arg("Nbeam", kwargs, 324)

        def __init__(self, **kwargs):
            self.compact = _spec_arg("compact", kwargs, self._base_compact())
            self.IDnorm = _spec_arg("IDnorm", kwargs, 0.5)
            self.ODnorm = _spec_arg("ODnorm", kwargs, 0.8)
            self.full = _spec_arg("full", kwargs, self._base_full())
            self.D = _spec_arg("D", kwargs, 0.0463)

    class _base_F3:
        class _base_compact:
            def __init__(self, **kwargs):
                self.dummy = _spec_arg("dummy", kwargs, 1)
                self.res = _spec_arg("res", kwargs, 30.0)

        class _base_full:
            def __init__(self, **kwargs):
                self.dummy = _spec_arg("dummy", kwargs, 1)
                self.res = _spec_arg("res", kwargs, 50.0)

        def __init__(self, **kwargs):
            self.compact = _spec_arg("compact", kwargs, self._base_compact())
            self.full = _spec_arg("full", kwargs, self._base_full())
            self.Rin = _spec_arg("Rin", kwargs, 2.8)
            self.Rout = _spec_arg("Rout", kwargs, inf)
            self.ang = _spec_arg("ang", kwargs, 180)

    class _base_Fend:
        class _base_compact:
            def __init__(self, **kwargs):
                self.dummy = _spec_arg("dummy", kwargs, 1)
#                 self.res = _spec_arg("res", kwargs, 3)

        class _base_full:
            def __init__(self, **kwargs):
                self.dummy = _spec_arg("dummy", kwargs, 1)
#                 self.res = _spec_arg("res", kwargs, 6)

        class _base_score:
            def __init__(self, **kwargs):
#                 self.Rin = _spec_arg("Rin", kwargs, 2.8)
                self.dummy = _spec_arg("dummy", kwargs, 1)
#                 self.Rout = _spec_arg("Rout", kwargs, 10)
#                 self.ang = _spec_arg("ang", kwargs, 180)

        class _base_corr:
            def __init__(self, **kwargs):
#                 self.Rin = _spec_arg("Rin", kwargs, 2.8)
                self.dummy = _spec_arg("dummy", kwargs, 1)
#                 self.Rout = _spec_arg("Rout", kwargs, 10)
#                 self.ang = _spec_arg("ang", kwargs, 180)

        def __init__(self, **kwargs):
            self.compact = _spec_arg("compact", kwargs, self._base_compact())
            self.FOV = _spec_arg("FOV", kwargs, 11)
            self.full = _spec_arg("full", kwargs, self._base_full())
            self.score = _spec_arg("score", kwargs, self._base_score())
            self.corr = _spec_arg("corr", kwargs, self._base_corr())
            self.sides = _spec_arg("sides", kwargs, "both")


    def __repr__(self):
        for k, v in self.__dict__:
            print('%s: '%(k), v)
        pass

    def printInfo(self, level=1):
        print('--------------------------\n')
        print('Number of keys: %d\n\n'%(len(self.__dict__.keys())))
        #for k in self.__dict__:
        #    if type(self.__dict__[k]) in [str, int, bool, float] or type(self.__dict__[k]) not np.array:
        #        print('%s: %s\n'%(k, str(type(self.__dict__[k])), str(self.__dict__[k])))
        #    else:
        #        print('%s: \ttype:%s value:\n'%(k, str(type(self.__dict__[k]))))
        
    def __str__(self):
        retstr = ''
        retstr += '--------------------------\n'
        retstr += ('Number of keys: %d\n\n'%(len(self.__dict__.keys())))
        #retsrt += '\n'
        for k in self.__dict__:
            if type(self.__dict__[k]) in [str, int, bool, float]:
                retstr += ('%s: \ttype:%s value: %s\n'%(k, str(type(self.__dict__[k])), str(self.__dict__[k])))
                #retstr += ('mp.%s\n'%(k))
            else:
                retstr += ('%s: \ttype:%s value:\n'%(k, str(type(self.__dict__[k]))))
                #retstr += ('mp.%s\n'%(k))
        retstr += '--------------------------\n'
        #retsrt += '\n'
        return retstr

    def __init__(self, **kwargs):
        

        import os 
        dir_path = os.path.dirname(os.path.realpath(__file__))
        print(dir_path)

#         self.compact = _spec_arg("compact", kwargs, self._base_compact())
#         self.full = _spec_arg("full", kwargs, self._base_full())

#         self.Nitr = _spec_arg("Nitr", kwargs, 10)
#         self.SPname = _spec_arg("SPname", kwargs, 0)
#         self.TToffset = _spec_arg("TToffset", kwargs, 1)
#         self.TrialNum = _spec_arg("TrialNum", kwargs, 1)
#         self.planetFlag = _spec_arg("planetFlag", kwargs, 0)
#         self.Nsbp = _spec_arg("Nsbp", kwargs, 1)
#         self.SeriesNum = _spec_arg("SeriesNum", kwargs, 1)
#         self.thput_radius = _spec_arg("thput_radius", kwargs, 0.7)
#         self.lambda0 = _spec_arg("lambda0", kwargs, 5.75e-07)
#         self.flagApod = _spec_arg("flagApod", kwargs, 0)
#         self.centering = _spec_arg("centering", kwargs, "pixel")
#         self.Nwpsbp = _spec_arg("Nwpsbp", kwargs, 1)
#         self.useGPU = _spec_arg("useGPU", kwargs, 0)
#         self.NlamForTT = _spec_arg("NlamForTT", kwargs, 1)
#         self.flagNewPSD = _spec_arg("flagNewPSD", kwargs, 0)
#         self.pup_strut_width = _spec_arg("pup_strut_width", kwargs, 0.0322)
        self.dm1 = _spec_arg("dm1", kwargs, self._base_dm1())
        self.dm2 = _spec_arg("dm2", kwargs, self._base_dm2())
        self.dm8 = _spec_arg("dm1", kwargs, self._base_dm8())
        self.dm9 = _spec_arg("dm1", kwargs, self._base_dm9())
#         self.whichPupil = _spec_arg("whichPupil", kwargs, "WFIRST20180103")
#         self.flagDM2stop = _spec_arg("flagDM2stop", kwargs, 0)
#         self.d_dm1_dm2 = _spec_arg("d_dm1_dm2", kwargs, 1)
        self.P2 = _spec_arg("P2", kwargs, self._base_P2())
        self.P3 = _spec_arg("P3", kwargs, self._base_P3())
        self.P1 = _spec_arg("P1", kwargs, self._base_P1())
        self.P4 = _spec_arg("P4", kwargs, self._base_P4())
#         self.relinItrVec = _spec_arg("relinItrVec", kwargs, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
#         self.flagParfor = _spec_arg("flagParfor", kwargs, 0)
#         self.coro = _spec_arg("coro", kwargs, "LC")
#         self.Ntt = _spec_arg("Ntt", kwargs, 1)
#         self.controller = _spec_arg("controller", kwargs, "EFC")
#         self.logGmin = _spec_arg("logGmin", kwargs, -6)
#         self.fracBW = _spec_arg("fracBW", kwargs, 0.01)
#         self.LS_strut_width = _spec_arg("LS_strut_width", kwargs, 0.038)
#         self.fl = _spec_arg("fl", kwargs, 1)
        self.F3 = _spec_arg("F3", kwargs, self._base_F3())
        self.Fend = _spec_arg("Fend", kwargs, self._base_Fend())
#         self.d_P2_dm1 = _spec_arg("d_P2_dm1", kwargs, 0)
#         self.FPMampFac = _spec_arg("FPMampFac", kwargs, 0)
#         self.flagDM1stop = _spec_arg("flagDM1stop", kwargs, 0)
#         self.thput_eval_y = _spec_arg("thput_eval_y", kwargs, 0)
#         self.thput_eval_x = _spec_arg("thput_eval_x", kwargs, 6)
#         self.WspatialDef = _spec_arg("WspatialDef", kwargs, [])




"""Module for thin film functions to make complex-valued HLC occulters."""
import numpy as np
from os.path import isfile
import os
import numpy as np

import falco
from falco.check import real_scalar, real_positive_scalar,\
                            real_nonnegative_scalar, scalar_integer,\
                            positive_scalar_integer, real_array,\
                            oneD_array, twoD_array, twoD_square_array


def calc_complex_occulter(lam, aoi, t_Ti, t_Ni_vec, t_PMGI_vec,
                                 d0, pol, flagOPD=False, SUBSTRATE='FS'):
    """
    Calculate the thin-film complex transmission and reflectance.
    
    Calculates the thin-film complex transmission and reflectance for the
    provided combinations of metal and dielectric thicknesses and list of
    wavelengths.

    Parameters
    ----------
    lam : float
        Wavelength in meters.
    aoi : flaot
        Angle of incidence in degrees.
    t_Ti : float
        Titanium thickness in meters. Titanium goes only between
        fused silica and nickel layers.
    t_Ni_vec : array_like
        1-D array of nickel thicknesses in meters. Nickel goes between
        titanium and PMGI layers.
    t_PMGI_vec : array_like
        1-D array of PMGI thicknesses in meters.
    d0 : float
        Reference height for all phase offsets. Must be larger than the stack
        of materials, not including the substrate. Units of meters.
    pol : {0, 1, 2}
        Polarization state to compute values for.
        0 for TE(s) polarization,
        1 for TM(p) polarization,
        2 for mean of s and p polarizations
    flagOPD : bool, optional
        Flag to use the OPD convention. The default is False.
    SUBSTRATE : str, optional
        Material to use as the substrate. The default is 'FS'.

    Returns
    -------
    tCoef : numpy ndarray
        2-D array of complex transmission amplitude values.
    rCoef : numpy ndarray
        2-D array of complex reflection amplitude values.
    """
    real_positive_scalar(lam, 'lam', TypeError)
    real_nonnegative_scalar(aoi, 'theta', TypeError)
    real_nonnegative_scalar(t_Ti, 't_Ti', TypeError)
    oneD_array(t_Ni_vec, 't_Ni_vec', ValueError)
    oneD_array(t_PMGI_vec, 't_PMGI_vec', ValueError)
    # if len(t_Ti) != len(t_Ni_vec) or len(t_Ni_vec) != len(t_PMGI_vec):
    #     raise ValueError('Ti, Ni, and PMGI thickness vectors must all ' +
    #                      'have same length.')
    scalar_integer(pol, 'pol', TypeError)
    
    lam_nm = lam * 1.0e9  # m --> nm
    lam_um = lam * 1.0e6  # m --> microns
    lam_um2 = lam_um * lam_um
    theta = aoi * (np.pi/180.)  # deg --> rad
    
    # Define Material Properties
    # ---------------------------------------------
    # Substrate properties
    if SUBSTRATE.upper() in ('FS', 'FUSEDSILICA'):
        A1 = 0.68374049400
        A2 = 0.42032361300
        A3 = 0.58502748000
        B1 = 0.00460352869
        B2 = 0.01339688560
        B3 = 64.49327320000
        n_substrate = np.sqrt(1 + A1*lam_um2/(lam_um2 - B1) +
                           A2*lam_um2/(lam_um2 - B2) +
                           A3*lam_um2/(lam_um2 - B3))

    elif SUBSTRATE.upper() in ('N-BK7', 'NBK7', 'BK7'):
        B1 = 1.03961212
        B2 = 0.231792344
        B3 = 1.01046945
        C1 = 0.00600069867
        C2 = 0.0200179144
        C3 = 103.560653
        n_substrate = np.sqrt(1 + (B1*lam_um2/(lam_um2 - C1)) +
                              (B2*lam_um2/(lam_um2 - C2)) +
                              (B3*lam_um2/(lam_um2 - C3)))
    
    # Dielectric properties
    npmgi = 1.524 + 5.176e-03/lam_um**2 + 2.105e-4/lam_um**4
    Ndiel = len(t_PMGI_vec)
    
    # Metal layer properties
    # Titanium base layer under the nickel
    Nmetal = len(t_Ni_vec)
    t_Ti_vec = t_Ti * np.ones(Nmetal)
    t_Ti_vec[np.asarray(t_Ni_vec) < 1e-10] = 0  # no Ti where no Ni
    # from D Moody
    titanium = np.array([
                        [397, 2.08, 2.95],
                        [413, 2.14, 2.98],
                        [431, 2.21, 3.01],
                        [451, 2.27, 3.04],
                        [471, 2.3, 3.1],
                        [496, 2.36, 3.19],
                        [521, 2.44, 3.2],
                        [549, 2.54, 3.43],
                        [582, 2.6, 3.58],
                        [617, 2.67, 3.74],
                        [659, 2.76, 3.84],
                        [704, 2.86, 3.96],
                        [756, 3.00, 4.01],
                        [821, 3.21, 4.01],
                        [892, 3.29, 3.96],
                        [984, 3.35, 3.97],
                        [1088, 3.5, 4.02],
                        [1216, 3.62, 4.15]
                        ])
    lam_ti = titanium[:, 0]  # nm
    n_ti = titanium[:, 1]
    k_ti = titanium[:, 2]
    nti = np.interp(lam_nm, lam_ti, n_ti)
    kti = np.interp(lam_nm, lam_ti, k_ti)
    
    # Nickel
    localpath = os.path.dirname(os.path.abspath(__file__))
    fnNickel = os.path.join(localpath, 'data',
                            'nickel_data_from_Palik_via_Bala_wvlNM_n_k.txt')
    vnickel = np.loadtxt(fnNickel, delimiter="\t", unpack=False, comments="#")
    lam_nickel = vnickel[:, 0]  # nm
    n_nickel = vnickel[:, 1]
    k_nickel = vnickel[:, 2]
    nnickel = np.interp(lam_nm, lam_nickel, n_nickel)
    knickel = np.interp(lam_nm, lam_nickel, k_nickel)
    
    # Compute the complex transmission
    # tCoef = np.zeros((Nmetal, ), dtype=complex)  # initialize
    # rCoef = np.zeros((Nmetal, ), dtype=complex)  # initialize

    # for ii in range(Nmetal):
    #     dni = t_Ni_vec[ii]
    #     dti = t_Ti_vec[ii]
    #     dpm = t_PMGI_vec[ii]
        
    #     nvec = np.array([1, 1, npmgi, nnickel-1j*knickel, nti-1j*kti,
    #                      n_substrate], dtype=complex)
    #     dvec = np.array([d0-dpm-dni-dti, dpm, dni, dti])
        
    #     # Choose polarization
    #     if(pol == 2):  # Mean of the two
    #         [dummy1, dummy2, rr0, tt0] = solver(nvec, dvec, theta,
    #                                             lam, False)
    #         [dummy1, dummy2, rr1, tt1] = solver(nvec, dvec, theta,
    #                                             lam, True)
    #         rr = (rr0+rr1)/2.
    #         tt = (tt0+tt1)/2.
    #     elif(pol == 0 or pol == 1):
    #         [dumm1, dummy2, rr, tt] = solver(nvec, dvec, theta, lam,
    #                                          bool(pol))
    #     else:
    #         raise ValueError('Wrong input value for polarization.')

    #     # Choose phase convention
    #     if not flagOPD:
    #         tCoef[ii] = np.conj(tt)  # Complex field transmission coef
    #         rCoef[ii] = np.conj(rr)  # Complex field reflection coef
    #     else:  # OPD phase convention
    #         tCoef[ii] = tt  # Complex field transmission coeffient
    #         rCoef[ii] = rr  # Complex field reflection coeffient
                
    # Compute the complex transmission
    tCoef = np.zeros((Ndiel, Nmetal), dtype=complex)  # initialize
    rCoef = np.zeros((Ndiel, Nmetal), dtype=complex)  # initialize
    for jj in range(Ndiel):
        dpm = t_PMGI_vec[jj]
        
        for ii in range(Nmetal):
            dni = t_Ni_vec[ii]
            dti = t_Ti_vec[ii]
            
            nvec = np.array([1, 1, npmgi, nnickel-1j*knickel, nti-1j*kti,
                              n_substrate], dtype=complex)
            dvec = np.array([d0-dpm-dni-dti, dpm, dni, dti])
            
            # Choose polarization
            if(pol == 2):  # Mean of the two
                [dummy1, dummy2, rr0, tt0] = solver(nvec, dvec, theta,
                                                    lam, False)
                [dummy1, dummy2, rr1, tt1] = solver(nvec, dvec, theta,
                                                    lam, True)
                rr = (rr0+rr1)/2.
                tt = (tt0+tt1)/2.
            elif(pol == 0 or pol == 1):
                [dumm1, dummy2, rr, tt] = solver(nvec, dvec, theta, lam,
                                                  bool(pol))
            else:
                raise ValueError('Wrong input value for polarization.')

            # Choose phase convention
            if not flagOPD:
                tCoef[jj, ii] = np.conj(tt)  # Complex field transmission coef
                rCoef[jj, ii] = np.conj(rr)  # Complex field reflection coef
            else:  # OPD phase convention
                tCoef[jj, ii] = tt  # Complex field transmission coeffient
                rCoef[jj, ii] = rr  # Complex field reflection coeffient
    
    return tCoef, rCoef


def solver(n, d0, theta, lam, tetm=False):
    """
    Solve the thin film equations for the given materials.

    Parameters
    ----------
    n : array_like
        index of refraction for each layer.
        n(1) = index of incident medium
        n(N) = index of transmission medium
        then length(n) must be >= 2
    d0 : array_like
        thickness of each layer, not counting incident medium or transmission
        medium. length(d) = length(n)-2.
    theta : float
        angle of incidence [radians].
    lam : float
        wavelength. units of lam must be same as d0.
    tetm : bool, optional
        False => TE, True => TM. The default is False.

    Returns
    -------
    R : numpy ndarray
        normalized reflected intensity coefficient
    T : numpy ndarray
        normalized transmitted intensity coefficient
    rr : numpy ndarray
        complex field reflection coefficient
    tt : numpy ndarray
        complex field transmission coefficient

    """
    oneD_array(n, 'n', ValueError)
    oneD_array(d0, 'd0', ValueError)
    N = len(n)
    if not (len(d0) == N-2):
        raise ValueError('n and d size mismatch')
        pass
    real_nonnegative_scalar(theta, 'theta', TypeError)
    real_positive_scalar(lam, 'lam', TypeError)
    if not type(tetm) == bool:
        raise TypeError('tetm must be a boolean.')
    
    # np.hstac
    d = np.hstack((0, d0.reshape(len(d0, )), 0))
        
    kx = 2*np.pi*n[0]*np.sin(theta)/lam
    # sign agrees with measurement convention:
    kz = -np.sqrt((2*np.pi*n/lam)**2 - kx*kx)
    
    if tetm:
        kzz = kz/(n*n)
    else:
        kzz = kz
    
    eep = np.exp(-1j*kz*d)
    eem = np.exp(1j*kz*d)
    
    i1 = np.arange(N-1)
    i2 = np.arange(1, N)
    tin = 0.5*(kzz[i1] + kzz[i2])/kzz[i1]
    ri = (kzz[i1] - kzz[i2])/(kzz[i1] + kzz[i2])
    
    A = np.eye(2, dtype=complex)
    for i in range(N-1):
        A = A @ np.array(tin[i]*np.array([[eep[i], ri[i]*eep[i]],
                                          [ri[i]*eem[i], eem[i]]]))
    
    rr = A[1, 0] / A[0, 0]
    tt = 1 / A[0, 0]
    
    # transmitted power flux (Poynting vector . surface) depends on index of
    # the substrate and angle
    R = np.abs(rr)**2
    if tetm:
        Pn = np.real((kz[-1]/(n[-1]**2)) / (kz[0]/(n[0]**2)))
    else:
        Pn = np.real((kz[-1]/kz[0]))
        pass
    
    T = Pn*np.abs(tt)**2
    tt = np.sqrt(Pn)*tt
    
    return [R, T, rr, tt]
   

def gen_complex_trans_table(mp, flagRefl=False, SUBSTRATE='FS'):
    """
    Calculate 3-D look-up table for thin film transmission data.
    
    Calculate thin-film complex transmission data cube. The three dimensions
    are for metal thickness, dielectric thickness, and wavelength.
    
    Parameters
    ----------
    mp : ModelParameters
        Model parameters object.
    flagRefl : TYPE, optional
        Compute the thin film properties in reflection. The default is False.
    SUBSTRATE : TYPE, optional
        Change the substrate material. The default is 'FS'. The only other
        option is 'BK7'.

    Returns
    -------
    complexTransCompact : numpy ndarray
        Complex transmission datacube for FALCO's compact model.
    complexTransFull : numpy ndarray
        Complex transmission datacube for FALCO's full model.
    """
    localpath = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    
    mp.F3.metal = 'Ni'
    mp.F3.diel = 'PMGI'
    
    fn_compact = ('ct_cube_%s_Ti%.1fnm_%s_%.1fto%.1fby%.2f_%s_%.1fto%.1fby%.2f_wvl%dnm_BW%.1fN%d_%.1fdeg_compact.npy' %
        (SUBSTRATE, mp.t_Ti_nm, mp.F3.metal, np.min(mp.t_metal_nm_vec),
         np.max(mp.t_metal_nm_vec), mp.dt_metal_nm, mp.F3.diel,
         np.min(mp.t_diel_nm_vec), np.max(mp.t_diel_nm_vec), mp.dt_diel_nm,
         1e9*mp.lambda0, 100*mp.fracBW, mp.Nsbp, mp.aoi))
    fn_cube_compact = os.path.join(localpath, 'data', 'material', fn_compact)
    
    fn_full = ('ct_cube_%s_Ti%.1fnm_%s_%.1fto%.1fby%.2f_%s_%.1fto%.1fby%.2f_wvl%dnm_BW%.1f_%dN%d_%.1fdeg_full.npy' %
        (SUBSTRATE, mp.t_Ti_nm, mp.F3.metal, np.min(mp.t_metal_nm_vec), np.max(mp.t_metal_nm_vec), mp.dt_metal_nm,
        mp.F3.diel, np.min(mp.t_diel_nm_vec), np.max(mp.t_diel_nm_vec), mp.dt_diel_nm,
        (1e9*mp.lambda0), 100*mp.fracBW, mp.Nsbp, mp.Nwpsbp, mp.aoi))
    fn_cube_full = os.path.join(localpath, 'data', 'material', fn_full)

    if(flagRefl):
        fn_cube_compact = fn_cube_compact[0:-4] + '_refl.mat'
        fn_cube_full = fn_cube_full[0:-4] + '_refl.mat'
    
    t_Ti_m = 1e-9*mp.t_Ti_nm  # Static base layer of titanium beneath nickel.
    aoi = mp.aoi
    Nsbp = mp.Nsbp
    t_diel_m_vec = 1e-9*mp.t_diel_nm_vec  # PMGI thickness range
    t_metal_m_vec = 1e-9*mp.t_metal_nm_vec  # nickel thickness range
    
    Nmetal = len(mp.t_metal_nm_vec)
    Ndiel = len(mp.t_diel_nm_vec)
    
    # Compact Model: Load the data if it has been generated before; otherwise generate it.
    if(isfile(fn_cube_compact)):
        complexTransCompact = np.load(fn_cube_compact)
        print('Loaded complex transmission datacube for compact model: %s' % fn_cube_compact)
    else:
    
        print('Computing thin film equations for compact model:')
        complexTransCompact = np.zeros((Ndiel, Nmetal, mp.Nsbp), dtype=complex)
        sbp_centers = mp.sbp_centers
        
        # Parallel/distributed computing
        # To be completed later
        
        # Regular (serial) computing
        for si in range(Nsbp):
            lam = sbp_centers[si]
            d0 = lam * mp.F3.d0fac  # Max thickness of PMGI + Ni
            [tCoef, rCoef] = calc_complex_occulter(lam, aoi, t_Ti_m,
                                            t_metal_m_vec, t_diel_m_vec, d0, 2)
            if(flagRefl):
                complexTransCompact[:, :, si] = rCoef
            else:
                complexTransCompact[:, :, si] = tCoef
                pass
            print('\tDone computing wavelength %d of %d.\n' % (si, Nsbp))

        # Save out for future use
        np.save(fn_cube_compact, complexTransCompact)
        print('Saved complex transmission datacube: %s' % fn_cube_compact)
        pass
    
    # Full Model: Load the data if it has been generated before; otherwise generate it.
    if isfile(fn_cube_full):
        complexTransFull = np.load(fn_cube_full)
        print('Loaded complex transmission datacube for full model: %s' % fn_cube_full)
    else:
        print('Computing thin film equations for full model:')
        if mp.Nwpsbp == 1:
            complexTransFull = complexTransCompact
        else:
            complexTransFull = np.zeros((Ndiel, Nmetal, mp.Nsbp*mp.Nwpsbp),
                                        dtype=complex)
            lambdas = mp.full.lambdas
            
            # Parallel/distributed computing
            # To be completed later
            
            # Regular (serial) computing
            for li in range(len(lambdas)):
                lam = lambdas[li]
                d0 = lam * mp.F3.d0fac  # Max thickness of PMGI + Ni
                [tCoef, rCoef] = calc_complex_occulter(lam, aoi, t_Ti_m,
                                            t_metal_m_vec, t_diel_m_vec, d0, 2)
                if(flagRefl):
                    complexTransFull[:, :, li] = rCoef
                else:
                    complexTransFull[:, :, li] = tCoef
                    pass
                print('\tDone computing wavelength %d of %d.\n' %
                      (li, len(lambdas)))
        
        # Save out for future use
        np.save(fn_cube_full, complexTransFull)
        print('Saved complex transmission datacube: %s\n' % fn_cube_full)
    
    return complexTransCompact, complexTransFull



import numpy as np
import logging
from falco import util, check
from falco.mask import falco_gen_vortex_mask
from scipy.signal import tukey

log = logging.getLogger(__name__)

_VALID_CENTERING = ['pixel', 'interpixel']
_CENTERING_ERR = 'Invalid centering specification. Options: {}'.format(_VALID_CENTERING)


def relay(E_in, Nrelay, centering='pixel'):
    """
    Perform re-imaging of the input E-field through optical relays.
    
    Propagate a field through Nrelay optical relays, without any intermediate
    mask multiplications. This results in a 180-degree rotation of the array 
    for each optical relay. Correct centering of the array must be maintained.

    Parameters
    ----------
    E_in : array_like
        Input electric field
    Nrelay: int
        Number of times to relay (and rotate by 180 degrees)
    centering : string
        Whether the input field is pixel-centered or inter-pixel-centered. If
        the array is pixel-centered, the output is shifted by 1 pixel in both
        axes after an odd number of relays.

    Returns
    -------
    E_out : array_like
        The output E-field. Same as the input E-field but rotated by 180
        degrees times the number of optical relays.

    """
    if centering not in _VALID_CENTERING:
        raise ValueError(_CENTERING_ERR)
    check.twoD_array(E_in, 'E_in', TypeError)
    check.scalar_integer(Nrelay, 'Nrelay', TypeError)

    #--Only rotate if odd number of 180-degree rotations. If even, no change.
    if(np.mod(Nrelay,2)==1):
        # Reverse and scale input to account for propagation
        E_out = E_in[::-1, ::-1]  
        if centering == 'pixel':
            # Move the DC pixel back to the right place
            E_out = np.roll(E_out, (1, 1), axis=(0, 1))  
    else:
        E_out = E_in
        
    return E_out


def ptp(E_in, full_width, wavelength, dz):
    """
    Propagate an electric field array using the angular spectrum technique.

    Parameters
    ----------
    E_in : array_like
        Square (i.e. NxN) input array.
    full_width : float
        The width along each side of the array [meters]
    wavelength : float
        Propagation wavelength [meters]
    dz : float
        Axial propagation distance [meters]

    Returns
    -------
    array_like
        Field after propagating over distance dz.

    """
    check.twoD_array(E_in, 'E_in', TypeError)
    check.real_positive_scalar(full_width, 'full_width', TypeError)
    check.real_positive_scalar(wavelength, 'wavelength', TypeError)
    check.real_scalar(dz, 'dz', TypeError)
    
    M, N = E_in.shape
    dx = full_width / N
    N_critical = int(np.floor(wavelength * np.abs(dz) / (dx ** 2)))  # Critical sampling

    if M != N:  # Input array is not square
        raise ValueError('Input array is not square')

    elif N < N_critical:
        log.warning(
             '''
             Input array is undersampled.
                Minimum required samples:  {}
                                  Actual:  {}
             '''.format(N_critical, N))

    fx = np.arange(-N // 2, N // 2) / full_width
    rho = util.radial_grid(fx)  # Spatial frequency coordinate grid

    kernel = np.fft.fftshift(np.exp(-1j * np.pi * wavelength * dz * (rho ** 2)))
    intermediate = np.fft.fftn(np.fft.fftshift(E_in))

    return np.fft.ifftshift(np.fft.ifftn(kernel * intermediate))


def mft_f2p(E_foc, fl, wavelength, dxi, deta, dx, N, centering='pixel'):
    """
    Propagate a field from a focal plane to a pupil plane, using a matrix-multiply DFT.

    Parameters
    ----------
    E_foc : array_like
        Electric field array in focal plane
    fl : float
        Focal length of Fourier transforming lens
    wavelength : float
        Propagation wavelength
    dxi : float
        Step size along horizontal axis of focal plane
    deta : float
        Step size along vertical axis of focal plane
    dx : float
        Step size along either axis of focal plane.  The vertical and horizontal step sizes are
        assumed to be equal.
    N : int
        Number of datapoints along each side of the pupil-plane (output) array
    centering : string
        Whether the input and output arrays are pixel-centered or inter-pixel-centered.
        Possible values: 'pixel', 'interpixel'

    Returns
    -------
    array_like
        Field in pupil plane, after propagating through Fourier transforming lens

    """
    if centering not in _VALID_CENTERING:
        raise ValueError(_CENTERING_ERR)
    check.twoD_array(E_foc, 'E_foc', TypeError)
    check.real_scalar(fl, 'fl', TypeError)
    check.real_positive_scalar(wavelength, 'wavelength', TypeError)
    check.real_positive_scalar(dxi, 'dxi', TypeError)
    check.real_positive_scalar(deta, 'deta', TypeError)
    check.real_positive_scalar(dx, 'dx', TypeError)
    check.positive_scalar_integer(N, 'N', TypeError)

    Neta, Nxi = E_foc.shape
    dy = dx  # Assume equal sample spacing along both directions

    # Focal-plane coordinates
    xi = util.create_axis(Nxi, dxi, centering=centering)[:, None]  # Broadcast to column vector
    eta = util.create_axis(Neta, dxi, centering=centering)[None, :]  # Broadcast to row vector

    # Pupil-plane coordinates
    x = util.create_axis(N, dx, centering=centering)[None, :]  # Broadcast to row vector
    y = x.T  # Column vector

    # Fourier transform matrices
    pre = np.exp(-2 * np.pi * 1j * (y * eta) / (wavelength * fl))
    post = np.exp(-2 * np.pi * 1j * (xi * x) / (wavelength * fl))

    # Constant scaling factor in front of Fourier transform
    scaling = np.sqrt(dx * dy * dxi * deta) / (1 * wavelength * fl)

    return scaling * np.linalg.multi_dot([pre, E_foc, post])


def mft_p2f(E_pup, fl, wavelength, dx, dxi, Nxi, deta, Neta, centering='pixel'):
    """
    Propagate a pupil to a focus using a matrix-multiply DFT.

    Parameters
    ----------
    E_pup : array_like
        Electric field array in pupil plane
    fl : float
        Focal length of Fourier transforming lens
    wavelength : float
        Propagation wavelength
    dx : float
        Step size along either axis of focal plane.  The vertical and horizontal step sizes are
        assumed to be equal.
    dxi : float
        Step size along horizontal axis of focal plane
    Nxi : int
        Number of samples along horizontal axis of focal plane.
    deta : float
        Step size along vertical axis of focal plane
    Neta : int
        Number of samples along vertical axis of focal plane.
    centering : string
        Whether the input and output arrays are pixel-centered or inter-pixel-centered.
        Possible values: 'pixel', 'interpixel'

    Returns
    -------
    array_like
        Field in pupil plane, after propagating through Fourier transforming lens

    """
    if centering not in _VALID_CENTERING:
        raise ValueError(_CENTERING_ERR)
    check.twoD_array(E_pup, 'E_pup', TypeError)
    check.real_scalar(fl, 'fl', TypeError)
    check.real_positive_scalar(wavelength, 'wavelength', TypeError)
    check.real_positive_scalar(dx, 'dx', TypeError)
    check.real_positive_scalar(dxi, 'dxi', TypeError)
    check.positive_scalar_integer(Nxi, 'Nxi', TypeError)
    check.real_positive_scalar(deta, 'deta', TypeError)
    check.positive_scalar_integer(Neta, 'Neta', TypeError)

    dy = dx
    M, N = E_pup.shape
    if M != N:
        raise ValueError('Input array is not square')

    # Pupil-plane coordinates
    x = util.create_axis(N, dx, centering=centering)[:, None]  # Broadcast to column vector
    y = x.T  # Row vector

    # Focal-plane coordinates
    xi = util.create_axis(Nxi, dxi, centering=centering)[None, :]  # Broadcast to row vector
    eta = util.create_axis(Neta, deta, centering=centering)[:, None]  # Broadcast to column vector

    # Fourier transform matrices
    pre = np.exp(-2 * np.pi * 1j * (eta * y) / (wavelength * fl))
    post = np.exp(-2 * np.pi * 1j * (x * xi) / (wavelength * fl))

    # Constant scaling factor in front of Fourier transform
    scaling = np.sqrt(dx * dy * dxi * deta) / (1 * wavelength * fl)

    return scaling * np.linalg.multi_dot([pre, E_pup, post])


def mft_p2v2p(pupilPre, charge, beamRadius, inVal, outVal):
    """
    Propagate from the pupil plane before a vortex FPM to pupil plane after it.
    
    Compute a radial Tukey window for propagating through a vortex coroangraph.

    Parameters
    ----------
    pupilPre : array_like
        2-D E-field at pupil plane before the vortex focal plane mask
    charge : int, float
        Charge of the vortex mask
    beamRadius : float
        Beam radius at pupil plane. Units of pixels.
    inVal : float
        Ask Gary
    outVal : float
        Ask Gary
        
    Returns
    -------
    pupilPost : array_like
        2-D E-field at pupil plane after the vortex focal plane mask

    """
    check.twoD_array(pupilPre, 'pupilPre', TypeError)
    check.scalar_integer(charge, 'charge', TypeError)
    check.real_positive_scalar(beamRadius, 'beamRadius', TypeError)
    check.real_positive_scalar(inVal, 'inVal', TypeError)
    check.real_positive_scalar(outVal, 'outVal', TypeError)
    
    # showPlots2debug = False 

    D = 2.0*beamRadius
    lambdaOverD = 4. # samples per lambda/D
    
    NA = pupilPre.shape[1]
    NB = util.ceil_even(lambdaOverD*D)
    
    # [X,Y] = np.meshgrid(np.arange(-NB/2., NB/2., dtype=float),np.arange(-NB/2., NB/2., dtype=float))
    # [RHO,THETA] = util.cart2pol(Y,X)
    RHO = util.radial_grid(np.arange(-NB/2., NB/2., dtype=float))
   
    windowKnee = 1.-inVal/outVal
    
    windowMask1 = gen_tukey_for_vortex(2*outVal*lambdaOverD, RHO, windowKnee)
    windowMask2 = gen_tukey_for_vortex(NB, RHO, windowKnee)

    # DFT vectors 
    x = np.arange(-NA/2,NA/2,dtype=float)/D   #(-NA/2:NA/2-1)/D
    u1 = np.arange(-NB/2,NB/2,dtype=float)/lambdaOverD #(-NB/2:NB/2-1)/lambdaOverD
    u2 = np.arange(-NB/2,NB/2,dtype=float)*2*outVal/NB # (-NB/2:NB/2-1)*2*outVal/N
    
    FPM = falco_gen_vortex_mask(charge, NB)

    #if showPlots2debug; figure;imagesc(abs(pupilPre));axis image;colorbar; title('pupil'); end;

    ## Low-sampled DFT of entire region

    FP1 = 1/(1*D*lambdaOverD)*np.exp(-1j*2*np.pi*np.outer(u1,x)) @ pupilPre @ np.exp(-1j*2*np.pi*np.outer(x,u1))
    #if showPlots2debug; figure;imagesc(log10(abs(FP1).^2));axis image;colorbar; title('Large scale DFT'); end;

    LP1 = 1/(1*D*lambdaOverD)*np.exp(-1j*2*np.pi*np.outer(x,u1)) @ (FP1*FPM*(1-windowMask1)) @ np.exp(-1j*2*np.pi*np.outer(u1,x))
    #if showPlots2debug; figure;imagesc(abs(FP1.*(1-windowMask1)));axis image;colorbar; title('Large scale DFT (windowed)'); end;
    
    ## Fine sampled DFT of innter region
    FP2 = 2*outVal/(1*D*NB)*np.exp(-1j*2*np.pi*np.outer(u2,x)) @ pupilPre @ np.exp(-1j*2*np.pi*np.outer(x,u2))
    #if showPlots2debug; figure;imagesc(log10(abs(FP2).^2));axis image;colorbar; title('Fine sampled DFT'); end;
    FPM = falco_gen_vortex_mask(charge, NB)
    LP2 = 2.0*outVal/(1*D*NB)*np.exp(-1j*2*np.pi*np.outer(x,u2)) @ (FP2*FPM*windowMask2) @ np.exp(-1j*2*np.pi*np.outer(u2,x))       
    #if showPlots2debug; figure;imagesc(abs(FP2.*windowMask2));axis image;colorbar; title('Fine sampled DFT (windowed)'); end;
    pupilPost = LP1 + LP2;
    #if showPlots2debug; figure;imagesc(abs(pupilPost));axis image;colorbar; title('Lyot plane'); end;

    return pupilPost


def gen_tukey_for_vortex(Nwindow, RHO, alpha):
    """
    Compute a radial Tukey window for propagating through a vortex coroangraph.

    Parameters
    ----------
    Nwindow : float, int
        Ask Gary
    RHO : array_like
        Radial coordinates over which to compute a Tukey function
    alpha : float
        Shape parameter of the Tukey window, representing the fraction of the
        window inside the cosine tapered region.

    Returns
    -------
    windowTukey : array_like
        Tukey window of same size as input RHO 

    """
    check.real_scalar(Nwindow, 'Nwindow', TypeError)
    check.real_scalar(alpha, 'alpha', TypeError)

    Nlut = int(10*Nwindow)
    rhos0 = np.linspace(-Nwindow/2, Nwindow/2, Nlut)
    lut = tukey(Nlut, alpha) #,left=0,right=0)
    windowTukey = np.interp(RHO, rhos0, lut)

    return windowTukey



"""Control functions for WFSC."""

import numpy as np
# import multiprocessing
# from astropy.io import fits
# import matplotlib.pyplot as plt
import falco


def wrapper(mp, cvar, jacStruct):
    """
    Outermost wrapper function for all the controller functions.

    Parameters
    ----------
    mp : ModelParameters
        Structure containing optical model parameters
    cvar : ModelParameters
        Structure containing controller variables
    jacStruct : ModelParameters
        Structure containing control Jacobians for each specified DM.

    Returns
    -------
    None
        Changes are made by reference to mp.
    """
#    if type(mp) is not falco.config.ModelParameters:
#        raise TypeError('Input "mp" must be of type ModelParameters')
#    pass

    # with falco.util.TicToc('Using the Jacobian to make other matrices'):
    print('Using the Jacobian to make other matrices...', end='')
    
    # Compute matrices for linear control with regular EFC
    cvar.GstarG_wsum = np.zeros((cvar.NeleAll, cvar.NeleAll))
    cvar.RealGstarEab_wsum = np.zeros((cvar.NeleAll, 1))

    for im in range(mp.jac.Nmode):

        Gmode = np.zeros((mp.Fend.corr.Npix, 1), dtype=complex)  # Initialize a row to concatenate onto
        if(any(mp.dm_ind == 1)): Gmode = np.hstack((Gmode,np.squeeze(jacStruct.G1[:,:,im])))
        if(any(mp.dm_ind == 2)): Gmode = np.hstack((Gmode,np.squeeze(jacStruct.G2[:,:,im])))
        if(any(mp.dm_ind == 8)): Gmode = np.hstack((Gmode,np.squeeze(jacStruct.G8[:,:,im])))
        if(any(mp.dm_ind == 9)): Gmode = np.hstack((Gmode,np.squeeze(jacStruct.G9[:,:,im])))
        Gmode = Gmode[:, 1:]  # Remove the zero column used for initialization

        # Square matrix part stays the same if no re-linearization has occurrred.
        cvar.GstarG_wsum += mp.jac.weights[im]*np.real(np.conj(Gmode).T @ Gmode)

        # The G^*E part changes each iteration because the E-field changes.
        # Apply 2-D spatial weighting to E-field in dark hole pixels.
        Eweighted = mp.WspatialVec*cvar.EfieldVec[:, im]
        # Apply the Jacobian weights and add to the total.
        cvar.RealGstarEab_wsum += mp.jac.weights[im]*np.real(
            np.conj(Gmode).T @ Eweighted.reshape(mp.Fend.corr.Npix, 1))
    
    # Make the regularization matrix. (Define only diagonal here to save RAM.)
    cvar.EyeGstarGdiag = np.max(np.diag(cvar.GstarG_wsum))*np.ones(cvar.NeleAll)
    cvar.EyeNorm = np.max(np.diag(cvar.GstarG_wsum))
    print('done.')

    # Call the Controller Function
    print('Control beginning ...')
    # Established, conventional controllers
    if mp.controller.lower() == 'plannedefc':
        dDM = _planned_efc(mp, cvar)
    elif mp.controller.lower() == 'gridsearchefc':
        dDM = _grid_search_efc(mp, cvar)
            
    # Update the DM commands by adding the delta control signal
    if(any(mp.dm_ind == 1)): mp.dm1.V += dDM.dDM1V
    if(any(mp.dm_ind == 2)): mp.dm2.V += dDM.dDM2V
    if(any(mp.dm_ind == 8)): mp.dm8.V += dDM.dDM8V
    if(any(mp.dm_ind == 9)): mp.dm9.V += dDM.dDM9V

    # Save the delta from the previous command
    if(any(mp.dm_ind == 1)): mp.dm1.dV = dDM.dDM1V
    if(any(mp.dm_ind == 2)): mp.dm2.dV = dDM.dDM2V
    if(any(mp.dm_ind == 8)): mp.dm8.dV = dDM.dDM8V
    if(any(mp.dm_ind == 9)): mp.dm9.dV = dDM.dDM9V


def cull_actuators(mp, cvar, jacStruct):
    """
    Remove weak actuators from the controlled set.

    Parameters
    ----------
    mp : ModelParameters
        Structure containing optical model parameters
    cvar : ModelParameters
        Structure containing controller variables
    jacStruct : ModelParameters
        Structure containing control Jacobians for each specified DM.

    Returns
    -------
    None
        Changes are made by reference to mp and jacStruct.
    """
    if type(mp) is not falco.config.ModelParameters:
        raise TypeError('Input "mp" must be of type ModelParameters')
    
    # Reduce the number of actuators used based on their relative strength
    # in the Jacobian
    if(cvar.flagCullAct and cvar.flagRelin):
        
        print('Weeding out weak actuators from the control Jacobian...')
        if(any(mp.dm_ind == 1)):
            G1intNorm = np.sum(np.mean(np.abs(jacStruct.G1)**2, axis=2), axis=0)
            G1intNorm = G1intNorm/np.max(G1intNorm)
            mp.dm1.act_ele = np.nonzero(G1intNorm >= 10**(mp.logGmin))[0]
            del G1intNorm
        if(any(mp.dm_ind == 2)):
            G2intNorm = np.sum(np.mean(np.abs(jacStruct.G2)**2, axis=2), axis=0)
            G2intNorm = G2intNorm/np.max(G2intNorm)
            mp.dm2.act_ele = np.nonzero(G2intNorm >= 10**(mp.logGmin))[0]
            del G2intNorm
        if(any(mp.dm_ind == 8)):
            G8intNorm = np.sum(np.mean(np.abs(jacStruct.G8)**2, axis=2), axis=0)
            G8intNorm = G8intNorm/np.max(G8intNorm)
            mp.dm8.act_ele = np.nonzero(G8intNorm >= 10**(mp.logGmin))[0]
            del G8intNorm
        if(any(mp.dm_ind == 9)):
            G9intNorm = np.sum(np.mean(np.abs(jacStruct.G9)**2, axis=2), axis=0)
            G9intNorm = G9intNorm/np.max(G9intNorm)
            mp.dm9.act_ele = np.nonzero(G9intNorm >= 10**(mp.logGmin))[0]
            del G9intNorm

        # Add back in all actuators that are tied (to make the tied actuator
        # logic easier)
        if(any(mp.dm_ind == 1)):
            for ti in range(mp.dm1.tied.shape[0]):
                if not (any(mp.dm1.act_ele == mp.dm1.tied[ti, 0])):
                    mp.dm1.act_ele = np.hstack([mp.dm1.act_ele, mp.dm1.tied[ti, 0]])
                if not (any(mp.dm1.act_ele == mp.dm1.tied[ti, 1])):
                    mp.dm1.act_ele = np.hstack([mp.dm1.act_ele, mp.dm1.tied[ti, 1]])
            # Need to sort for the logic in model_Jacobian.m
            mp.dm1.act_ele = np.sort(mp.dm1.act_ele)

        if(any(mp.dm_ind == 2)):
            for ti in range(mp.dm2.tied.shape[0]):
                if not any(mp.dm2.act_ele == mp.dm2.tied[ti, 0]):
                    mp.dm2.act_ele = np.hstack([mp.dm2.act_ele, mp.dm2.tied[ti, 0]])
                if not any(mp.dm2.act_ele == mp.dm2.tied[ti, 1]):
                    mp.dm2.act_ele = np.hstack([mp.dm2.act_ele, mp.dm2.tied[ti, 1]])
            # Need to sort for the logic in model_Jacobian.m
            mp.dm2.act_ele = np.sort(mp.dm2.act_ele)
#            if(any(mp.dm_ind == 8))
#                for ti=1:size(mp.dm8.tied,1)
#                    if(any(mp.dm8.act_ele==mp.dm8.tied(ti,1))==false);  mp.dm8.act_ele = [mp.dm8.act_ele; mp.dm8.tied(ti,1)];  end
#                    if(any(mp.dm8.act_ele==mp.dm8.tied(ti,2))==false);  mp.dm8.act_ele = [mp.dm8.act_ele; mp.dm8.tied(ti,2)];  end
#                end
#                mp.dm8.act_ele = sort(mp.dm8.act_ele);
#            end
#            if(any(mp.dm_ind == 9))
#                for ti=1:size(mp.dm9.tied,1)
#                    if(any(mp.dm9.act_ele==mp.dm9.tied(ti,1))==false);  mp.dm9.act_ele = [mp.dm9.act_ele; mp.dm9.tied(ti,1)];  end
#                    if(any(mp.dm9.act_ele==mp.dm9.tied(ti,2))==false);  mp.dm9.act_ele = [mp.dm9.act_ele; mp.dm9.tied(ti,2)];  end
#                end
#                mp.dm9.act_ele = sort(mp.dm9.act_ele);
#            end
            
        # Update the number of elements used per DM
        if(any(mp.dm_ind == 1)): mp.dm1.Nele = mp.dm1.act_ele.size
        if(any(mp.dm_ind == 2)): mp.dm2.Nele = mp.dm2.act_ele.size
        if(any(mp.dm_ind == 8)): mp.dm8.Nele = mp.dm8.act_ele.size
        if(any(mp.dm_ind == 9)): mp.dm9.Nele = mp.dm9.act_ele.size

        if(any(mp.dm_ind == 1)): print('  DM1: %d/%d (%.2f%%) actuators kept for Jacobian' % (mp.dm1.Nele, mp.dm1.NactTotal, 100*mp.dm1.Nele/mp.dm1.NactTotal))
        if(any(mp.dm_ind == 2)): print('  DM2: %d/%d (%.2f%%) actuators kept for Jacobian' % (mp.dm2.Nele, mp.dm2.NactTotal, 100*mp.dm2.Nele/mp.dm2.NactTotal))
        if(any(mp.dm_ind == 8)): print('  DM8: %d/%d (%.2f%%) actuators kept for Jacobian' % (mp.dm8.Nele, mp.dm8.NactTotal, 100*mp.dm8.Nele/mp.dm8.NactTotal))
        if(any(mp.dm_ind == 9)): print('  DM9: %d/%d (%.2f%%) actuators kept for Jacobian' % (mp.dm9.Nele, mp.dm9.NactTotal, 100*mp.dm9.Nele/mp.dm9.NactTotal))
        
        # Crop out unused actuators from the control Jacobian
        if(any(mp.dm_ind == 1)): jacStruct.G1 = jacStruct.G1[:, mp.dm1.act_ele, :]
        if(any(mp.dm_ind == 2)): jacStruct.G2 = jacStruct.G2[:, mp.dm2.act_ele, :]
        if(any(mp.dm_ind == 8)): jacStruct.G8 = jacStruct.G8[:, mp.dm8.act_ele, :]
        if(any(mp.dm_ind == 9)): jacStruct.G9 = jacStruct.G9[:, mp.dm9.act_ele, :]

    return None


def _grid_search_efc(mp, cvar):
    """
    Perform a grid search over specified variables for the controller.

    Parameters
    ----------
    mp : ModelParameters
        Structure containing optical model parameters
    cvar : ModelParameters
        Structure containing controller variables

    Returns
    -------
    dDM : ModelParameters
        Structure containing the delta DM commands for each DM
    """
    # Make all combinations of the values
    vals_list = [(x, y) for y in mp.ctrl.dmfacVec for x in mp.ctrl.log10regVec]
    Nvals = len(mp.ctrl.log10regVec) * len(mp.ctrl.dmfacVec)
    InormVec = np.zeros(Nvals)  # Initialize

    # Temporarily store computed DM commands so that the best one does not have
    # to be re-computed
    if(any(mp.dm_ind == 1)):
        dDM1V_store = np.zeros((mp.dm1.Nact, mp.dm1.Nact, Nvals))
    if(any(mp.dm_ind == 2)):
        dDM2V_store = np.zeros((mp.dm2.Nact, mp.dm2.Nact, Nvals))
    if(any(mp.dm_ind == 8)):
        dDM8V_store = np.zeros((mp.dm8.NactTotal, Nvals))
    if(any(mp.dm_ind == 9)):
        dDM9V_store = np.zeros((mp.dm9.NactTotal, Nvals))

    # Empirically find the regularization value giving the best contrast
#    if(mp.flagMultiproc and mp.ctrl.flagUseModel):
#        # Run the controller in parallel
#        pool = multiprocessing.Pool(processes=mp.Nthreads)
#        results = [pool.apply_async(_efc, args=(ni,vals_list,mp,cvar)) for ni in np.arange(Nvals,dtype=int) ]
#        results_ctrl = [p.get() for p in results] # All the Jacobians in a list
#        pool.close()
#        pool.join()
#
#        # Convert from a list to arrays:
#        for ni in range(Nvals):
#            InormVec[ni] = results_ctrl[ni][0]
#            if(any(mp.dm_ind == 1)): dDM1V_store[:,:,ni] = results_ctrl[ni][1].dDM1V
#            if(any(mp.dm_ind == 2)): dDM2V_store[:,:,ni] = results_ctrl[ni][1].dDM2V
#    else:
    for ni in range(Nvals):
        [InormVec[ni], dDM_temp] = _efc(ni, vals_list, mp, cvar)
        # delta voltage commands
        if(any(mp.dm_ind == 1)): dDM1V_store[:, :, ni] = dDM_temp.dDM1V
        if(any(mp.dm_ind == 2)): dDM2V_store[:, :, ni] = dDM_temp.dDM2V
        if(any(mp.dm_ind == 8)): dDM8V_store[:, ni] = dDM_temp.dDM8V
        if(any(mp.dm_ind == 9)): dDM9V_store[:, ni] = dDM_temp.dDM9V

    # Print out results to the command line
    print('Scaling factor:\t', end='')
    for ni in range(Nvals):
        print('%.2f\t\t' % (vals_list[ni][1]), end='')

    print('\nlog10reg:    \t', end='')
    for ni in range(Nvals):
        print('%.1f\t\t' % (vals_list[ni][0]), end='')
    
    print('\nInorm:       \t', end='')
    for ni in range(Nvals):
        print('%.2e\t' % (InormVec[ni]), end='')
    print('\n', end='')

    # Find the best scaling factor and Lagrange multiplier pair based on the
    # best contrast.
    # [cvar.cMin,indBest] = np.min(InormVec)
    indBest = np.argmin(InormVec)
    cvar.cMin = np.min(InormVec)
    dDM = falco.config.Object()
    # delta voltage commands
    if(any(mp.dm_ind == 1)): dDM.dDM1V = np.squeeze(dDM1V_store[:, :, indBest])
    if(any(mp.dm_ind == 2)): dDM.dDM2V = np.squeeze(dDM2V_store[:, :, indBest])
    if(any(mp.dm_ind == 8)): dDM.dDM8V = np.squeeze(dDM8V_store[:, indBest])
    if(any(mp.dm_ind == 9)): dDM.dDM9V = np.squeeze(dDM9V_store[:, indBest])

    cvar.log10regUsed = vals_list[indBest][0]
    dmfacBest = vals_list[indBest][1]
    if(mp.ctrl.flagUseModel):
        print('Model-based grid search expects log10reg, = %.1f,\t dmfac = %.2f,\t %4.2e normalized intensity.'
              % (cvar.log10regUsed, dmfacBest, cvar.cMin))
    else:
        print('Empirical grid search finds log10reg, = %.1f,\t dmfac = %.2f,\t %4.2e normalized intensity.'
              % (cvar.log10regUsed, dmfacBest, cvar.cMin))
    
    return dDM


def _planned_efc(mp, cvar):
    """
    Perform a scheduled/planned set of EFC iterations.

    Parameters
    ----------
    mp : ModelParameters
        Structure containing optical model parameters
    cvar : ModelParameters
        Structure containing controller variables

    Returns
    -------
    dDM : ModelParameters
        Structure containing the delta DM commands for each DM
    """
    # Make all combinations of the values
    vals_list = [(x, y) for y in mp.ctrl.dmfacVec for x in mp.ctrl.log10regVec]
    Nvals = len(mp.ctrl.log10regVec) * len(mp.ctrl.dmfacVec)
    InormVec = np.zeros(Nvals)  # Initialize

    # Make more obvious names for conditions:
    relinearizeNow = any(np.array(mp.gridSearchItrVec) == cvar.Itr)
    useBestLog10Reg = np.imag(mp.ctrl.log10regSchedIn[cvar.Itr]) != 0
    realLog10RegIsZero = np.real(mp.ctrl.log10regSchedIn[cvar.Itr]) == 0

    # Step 1: Empirically find the "optimal" regularization value
    # (if told to for this iteration).
    
    if relinearizeNow:
        
        # Temporarily store computed DM commands so that the best one does
        # not have to be re-computed
        if(any(mp.dm_ind == 1)):
            dDM1V_store = np.zeros((mp.dm1.Nact, mp.dm1.Nact, Nvals))
        if(any(mp.dm_ind == 2)):
            dDM2V_store = np.zeros((mp.dm2.Nact, mp.dm2.Nact, Nvals))
        if(any(mp.dm_ind == 8)):
            dDM8V_store = np.zeros((mp.dm8.NactTotal, Nvals))
        if(any(mp.dm_ind == 9)):
            dDM9V_store = np.zeros((mp.dm9.NactTotal, Nvals))
        
        for ni in range(Nvals):
            [InormVec[ni], dDM_temp] = _efc(ni, vals_list, mp, cvar)
            # delta voltage commands
            if(any(mp.dm_ind == 1)): dDM1V_store[:, :, ni] = dDM_temp.dDM1V
            if(any(mp.dm_ind == 2)): dDM2V_store[:, :, ni] = dDM_temp.dDM2V
            if(any(mp.dm_ind == 8)): dDM8V_store[:, ni] = dDM_temp.dDM8V
            if(any(mp.dm_ind == 9)): dDM9V_store[:, ni] = dDM_temp.dDM9V

        # Print out results to the command line
        print('Scaling factor:\t', end='')
        for ni in range(Nvals): print('%.2f\t\t' % (vals_list[ni][1]), end='')
    
        print('\nlog10reg:    \t', end='')
        for ni in range(Nvals): print('%.1f\t\t' % (vals_list[ni][0]), end='')
        
        print('\nInorm:       \t', end='')
        for ni in range(Nvals): print('%.2e\t' % (InormVec[ni]), end='')
        print('\n', end='')
    
        # Find the best scaling factor and Lagrange multiplier pair based on
        # the best contrast.
        # [cvar.cMin,indBest] = np.min(InormVec)
        indBest = np.argmin(InormVec)
        cvar.cMin = np.min(InormVec)
        cvar.latestBestlog10reg = vals_list[indBest][0]
        cvar.latestBestDMfac = vals_list[indBest][1]
        
        if(mp.ctrl.flagUseModel):
            print(('Model-based grid search expects log10reg, = %.1f,\t ' +
                  'dmfac = %.2f,\t %4.2e normalized intensity.') %
                  (cvar.latestBestlog10reg, cvar.latestBestDMfac, cvar.cMin))
        else:
            print(('Empirical grid search finds log10reg, = %.1f,\t dmfac' +
                  ' = %.2f,\t %4.2e normalized intensity.') %
                  (cvar.latestBestlog10reg, cvar.latestBestDMfac, cvar.cMin))
    
    # Skip steps 2 and 3 if the schedule for this iteration is just to use the
    # "optimal" regularization AND if grid search was performed this iteration.
    if relinearizeNow and useBestLog10Reg and realLog10RegIsZero:
        # delta voltage commands
        dDM = falco.config.Object()  # Initialize
        if(any(mp.dm_ind == 1)): dDM.dDM1V = np.squeeze(dDM1V_store[:, :, indBest])
        if(any(mp.dm_ind == 2)): dDM.dDM2V = np.squeeze(dDM2V_store[:, :, indBest])
        if(any(mp.dm_ind == 8)): dDM.dDM8V = np.squeeze(dDM8V_store[:, indBest])
        if(any(mp.dm_ind == 9)): dDM.dDM9V = np.squeeze(dDM9V_store[:, indBest])
        
        log10regSchedOut = cvar.latestBestlog10reg
    else:
        # Step 2: For this iteration in the schedule, replace the imaginary
        # part of the regularization with the latest "optimal" regularization
        if useBestLog10Reg:
            log10regSchedOut = cvar.latestBestlog10reg + \
                np.real(mp.ctrl.log10regSchedIn[cvar.Itr])
        else:
            log10regSchedOut = np.real(mp.ctrl.log10regSchedIn[cvar.Itr])

        # Step 3: Compute the EFC command to use
        ni = 0
        if not hasattr(cvar, 'latestBestDMfac'):
            cvar.latestBestDMfac = 1
        vals_list = [(x, y) for y in np.array([cvar.latestBestDMfac])
                      for x in np.array([log10regSchedOut])]
    
        [cvar.cMin, dDM] = _efc(ni, vals_list, mp, cvar)
        if mp.ctrl.flagUseModel:
            print(('Model expects scheduled log10(reg) = %.1f\t to give ' +
                  '%4.2e normalized intensity.') %
                  (log10regSchedOut, cvar.cMin))
        else:
            print(('Scheduled log10reg = %.1f\t gives %4.2e normalized' +
                  ' intensity.') % (log10regSchedOut, cvar.cMin))
    
    cvar.log10regUsed = log10regSchedOut
    
    return dDM


def efc_schedule_generator(scheduleMatrix):
    """
    Generate the EFC schedule from an input matrix.

    Parameters
    ----------
    scheduleMatrix : array_like
        DESCRIPTION.

    Returns
    -------
    Nitr : int
        Number of WFSC iterations.
    relinItrVec : array_like
        DESCRIPTION.
    gridSearchItrVec : array_like
        DESCRIPTION.
    log10regSched : array_like
        DESCRIPTION.
    dm_ind_sched : list
        DESCRIPTION.

    Notes
    -----
    CONTROL SCHEDULE. Columns of sched_mat are:
    % Column 0: # of iterations,
    % Column 1: log10(regularization),
    % Column 2: which DMs to use (12, 128, 129, or 1289) for control
    % Column 3: flag (0 = false, 1 = true), whether to re-linearize
    %   at that iteration.
    % Column 4: flag (0 = false, 1 = true), whether to perform an
    %   EFC parameter grid search to find the set giving the best
    %   contrast .
    % The imaginary part of the log10(regularization) in column 1 is
    %  replaced for that iteration with the optimal log10(regularization)
    % A row starting with [0, 0, 0, 1...] relinearizes only at that time
    """
    # Number of correction iterations
    Nitr = int(np.real(np.sum(scheduleMatrix[:, 0])))
    
    # Create the vectors of:
    #  1) iteration numbers at which to relinearize the Jacobian
    #  2) log10(regularization) at each correction iteration
    relinItrVec = np.array([])  # Initialize
    gridSearchItrVec = []  # Initialize
    log10regSched = np.zeros((Nitr,), dtype=complex)  # Initialize
    dmIndList = np.zeros((Nitr,), dtype=int)  # Initialize
    iterCount = 0
    for iRow in range(scheduleMatrix.shape[0]):
        
        # When to re-linearize
        if int(np.real(scheduleMatrix[iRow, 3])) == 1:
            relinItrVec = np.append(relinItrVec, iterCount)

        # When to re-do the empirical EFC grid search
        if int(np.real(scheduleMatrix[iRow, 4])) == 1:
            gridSearchItrVec.append(iterCount)
                
        # Make the vector of regularizations at each iteration
        deltaIter = int(np.real(scheduleMatrix[iRow, 0]))
        if not deltaIter == 0:
            log10regSched[iterCount:(iterCount+deltaIter)] = \
                scheduleMatrix[iRow, 1]
            dmIndList[iterCount:(iterCount+deltaIter)] = \
                int(np.real(scheduleMatrix[iRow, 2]))

        iterCount += int(np.real(scheduleMatrix[iRow, 0]))

    gridSearchItrVec = np.asarray(gridSearchItrVec)

    # Store DM number index vectors as cells since they can vary in length
    dm_ind_sched = []
    for Itr in range(Nitr):
        dm_ind = []
        numAsStr = str(dmIndList[Itr])
        for ii in range(len(numAsStr)):
            dm_ind.append(int(numAsStr[ii]))
        dm_ind_sched.append(np.array(dm_ind))

    return Nitr, relinItrVec, gridSearchItrVec, log10regSched, dm_ind_sched


def _efc(ni, vals_list, mp, cvar):
    """
    Compute the main EFC equation. Called by a wrapper controller function.

    Parameters
    ----------
    ni : int
        index for the set of possible combinations of variables to do a grid
        search over
    vals_list : list
        the set of possible combinations of values to do a grid search over
    mp : ModelParameters
        Structure containing optical model parameters
    cvar : ModelParameters
        Structure containing controller variables

    Returns
    -------
    InormAvg : float
        Normalized intensity averaged spectrally and spatially.
    dDM : ModelParameters
        Structure containing the delta DM commands for each DM
    """
    #  function [InormAvg,dDM] = _efc(ni,vals_list,mp,cvar)

    if(any(mp.dm_ind == 1)):
        DM1V0 = mp.dm1.V.copy()
    if(any(mp.dm_ind == 2)):
        DM2V0 = mp.dm2.V.copy()
    if(any(mp.dm_ind == 8)):
        DM8V0 = mp.dm8.V.copy()
    if(any(mp.dm_ind == 9)):
        DM9V0 = mp.dm9.V.copy()
        
    # Initializations
    log10reg = vals_list[ni][0]  # log 10 of regularization value
    dmfac = vals_list[ni][1]  # Scaling factor for entire DM command
    
    # Save starting point for each delta command to be added to.
    # Get the indices of each DM's command vector within the single
    # concatenated command vector
    init(mp, cvar)  # Modifies cvar
    
    # Least-squares solution with regularization:
    duVecNby1 = \
        -dmfac*np.linalg.solve((10**log10reg*np.diag(cvar.EyeGstarGdiag) +
                                cvar.GstarG_wsum), cvar.RealGstarEab_wsum)
    # Convert to true 1-D array from a 2-D, Nx1 array
    duVec = duVecNby1.reshape((-1,))
     
    # Parse the command vector by DM and assign the output commands
    mp, dDM = wrapup(mp, cvar, duVec)
    
    # Take images and compute average intensity in dark hole
    if(mp.ctrl.flagUseModel):  # Get simulated image from compact model
        Itotal = falco.imaging.get_expected_summed_image(mp, cvar, dDM)
        InormAvg = np.mean(Itotal[mp.Fend.corr.maskBool])
    else:  # Get actual image from full model or testbed
        Itotal = falco.imaging.get_summed_image(mp)
        InormAvg = np.mean(Itotal[mp.Fend.corr.maskBool])
        
    # Reset voltage commands in mp
    if(any(mp.dm_ind == 1)):
        mp.dm1.V = DM1V0
    if(any(mp.dm_ind == 2)):
        mp.dm2.V = DM2V0
    if(any(mp.dm_ind == 8)):
        mp.dm8.V = DM8V0
    if(any(mp.dm_ind == 9)):
        mp.dm9.V = DM9V0
        
    return InormAvg, dDM


def init(mp, cvar):
    """
    Vectorize DM commands and otherwise prepare variables for the controller.

    Parameters
    ----------
    mp : ModelParameters
        Structure containing optical model parameters
    cvar : ModelParameters
        Structure containing controller variables

    Returns
    -------
    None
        Changes are made by reference to structures mp and cvar.
    """
    # Save starting point for each delta command to be added to.
    if(any(mp.dm_ind == 1)): cvar.DM1Vnom = mp.dm1.V
    if(any(mp.dm_ind == 2)): cvar.DM2Vnom = mp.dm2.V
    if(any(mp.dm_ind == 8)): cvar.DM8Vnom = mp.dm8.V
    if(any(mp.dm_ind == 9)): cvar.DM9Vnom = mp.dm9.V
    
    # Make the vector of total DM commands from before
    u1 = mp.dm1.V.reshape(mp.dm1.NactTotal)[mp.dm1.act_ele] if(any(mp.dm_ind == 1)) else np.array([])
    u2 = mp.dm2.V.reshape(mp.dm2.NactTotal)[mp.dm2.act_ele] if(any(mp.dm_ind == 2)) else np.array([])
    u8 = mp.dm9.V.reshape(mp.dm8.NactTotal)[mp.dm8.act_ele] if(any(mp.dm_ind == 8)) else np.array([])
    u9 = mp.dm9.V.reshape(mp.dm9.NactTotal)[mp.dm9.act_ele] if(any(mp.dm_ind == 9)) else np.array([])
    cvar.uVec = np.concatenate((u1, u2, u8, u9))
    cvar.NeleAll = cvar.uVec.size
    
    # Get the indices of each DM's command within the full command
    u1dummy = 1*np.ones(mp.dm1.Nele, dtype=int) if(any(mp.dm_ind == 1)) else np.array([])
    u2dummy = 2*np.ones(mp.dm2.Nele, dtype=int) if(any(mp.dm_ind == 2)) else np.array([])
    u8dummy = 8*np.ones(mp.dm8.Nele, dtype=int) if(any(mp.dm_ind == 8)) else np.array([])
    u9dummy = 9*np.ones(mp.dm9.Nele, dtype=int) if(any(mp.dm_ind == 9)) else np.array([])
    cvar.uLegend = np.concatenate((u1dummy, u2dummy, u8dummy, u9dummy))
    

def wrapup(mp, cvar, duVec):
    """
    Apply controller commands to the DM command variables.

    Parameters
    ----------
    mp : ModelParameters
        Structure containing optical model parameters
    cvar : ModelParameters
        Structure containing controller variables
    duVec : numpy ndarray
        Vector of delta control commands computed by the controller.

    Returns
    -------
    mp : ModelParameters
        Structure containing optical model parameters
    dDM : ModelParameters
        Structure containing the delta DM commands for each DM
    """
    dDM = falco.config.Object()
    
    # Parse the command vector by DM and apply weighting
    if(any(mp.dm_ind == 1)):
        dDM1Vvec = np.zeros(mp.dm1.NactTotal)
        dDM1Vvec[mp.dm1.act_ele] = mp.dm1.weight*duVec[cvar.uLegend == 1]
        dDM.dDM1V = dDM1Vvec.reshape((mp.dm1.Nact, mp.dm1.Nact))
    if(any(mp.dm_ind == 2)):
        dDM2Vvec = np.zeros(mp.dm2.NactTotal)
        dDM2Vvec[mp.dm2.act_ele] = mp.dm2.weight*duVec[cvar.uLegend == 2]
        dDM.dDM2V = dDM2Vvec.reshape((mp.dm2.Nact, mp.dm2.Nact))
    if(any(mp.dm_ind == 8)):
        dDM.dDM8V = np.zeros(mp.dm8.NactTotal)
        dDM.dDM8V[mp.dm8.act_ele] = mp.dm8.weight*duVec[cvar.uLegend == 8]
    if(any(mp.dm_ind == 9)):
        dDM.dDM9V = np.zeros(mp.dm9.NactTotal)
        dDM.dDM9V[mp.dm9.act_ele] = mp.dm9.weight*duVec[cvar.uLegend == 9]
    
    # Enforce tied actuator pair commands.
    #   Assign command of first actuator to the second as well.
#    if(any(mp.dm_ind == 8)):
#        for ti=1:size(mp.dm8.tied,1)
#            dDM.dDM8V(mp.dm8.tied(ti,2)) = dDM.dDM8V(mp.dm8.tied(ti,1));
#
#    if(any(mp.dm_ind == 9)):
#        for ti=1:size(mp.dm9.tied,1)
#            dDM.dDM9V(mp.dm9.tied(ti,2)) = dDM.dDM9V(mp.dm9.tied(ti,1));

    # Combine the delta command with the previous command
    if(any(mp.dm_ind == 1)): mp.dm1.V = cvar.DM1Vnom + dDM.dDM1V
    if(any(mp.dm_ind == 2)): mp.dm2.V = cvar.DM2Vnom + dDM.dDM2V
    if(any(mp.dm_ind == 8)): mp.dm8.V = cvar.DM8Vnom + dDM.dDM8V
    if(any(mp.dm_ind == 9)): mp.dm9.V = cvar.DM9Vnom + dDM.dDM9V

    return mp, dDM
